#include "Tools.h"

const float PI = acos(-1.0f);

vector<QVector3D> createSubImage(
	const int &x, const int &y,
	const int &width, const int &height,
	const QImage &orig)
{
	vector<QVector3D> pixels;

	for(int i = x; i < width + x; i++)
	{
		for(int j = y; j < height + y; j++)
		{
			QRgb rgbValue = orig.pixel(i, j);
			pixels.push_back(
				QVector3D(qRed(rgbValue), qGreen(rgbValue), qBlue(rgbValue)));
		}
	}

	return pixels;
}

QImage createSubImage(
	const QImage &image, const QRect & rect)
{
	size_t offset = rect.x() * image.depth() / 8
		+ rect.y() * image.bytesPerLine();
	return QImage(image.bits() + offset, rect.width(), rect.height(),
		image.bytesPerLine(), image.format());
}


QImage flipImageHorizontally(const QImage &image)
{
	QImage imageSwapped(image.width(), image.height(), image.format());

	for(int x = 0; x < image.width(); x++)
	{
		for(int y = 1; y <= image.height(); y++)
		{
			imageSwapped.setPixel(x, image.width() - y,
				image.pixel(x, y - 1));
		}
	}

	return imageSwapped;
}

void createSubImage(
	const int &x, const int &y,
	const int &width, const int &height,
	const QVector2D &gradient, const QImage &orig,
	vector<QVector3D> &subImage1, vector<QVector3D> &subImage2)
{
	QVector2D separatingVector(gradient.y(), gradient.x());

	if( (x - width/2) < 0 || (x + 1 + width/2) > orig.width())
		return;
	if( (y - height/2) < 0 || (x + 1 + height/2) > orig.height())
		return;

	for(int i = -width/2; i < 1 + width/2; i++)
	{
		for(int j = -height/2; j < 1 + height/2; j++)
		{
			QRgb rgbValue = orig.pixel(
				x + i, y + j);
			if(QVector2D::dotProduct(separatingVector, QVector2D(i, j)) < 0)
				subImage1.push_back(
				QVector3D(qRed(rgbValue), qGreen(rgbValue), qBlue(rgbValue)));
			else
				subImage2.push_back(
				QVector3D(qRed(rgbValue), qGreen(rgbValue), qBlue(rgbValue)));
		}
	}
}

void drawSubImage(int x, int y, const QImage &subImage, QImage &image)
{
	for(int sX = -subImage.width()/2; sX < subImage.width()/2; sX++)
		for(int sY = -subImage.height()/2; sY < subImage.height()/2; sY++)
		{
			image.setPixel(x + sX, y + sY, subImage.pixel(
				sX + subImage.width()/2,
				sY + subImage.height()/2));
		}
}

void drawSubImage(int x, int y, int sideSize, const QColor &color, QImage &image)
{
	for(int sX = -sideSize/2; sX < sideSize/2; sX++)
	{
		for(int sY = -sideSize/2; sY < sideSize/2; sY++)
		{
			image.setPixel(x + sX, y + sY, color.rgba());
		}
	}
}

void drawCross(int x, int y, int sideSize, const QColor &color, QImage &image)
{
	for(int sX = -sideSize/2; sX < sideSize/2; sX++)
	{
		image.setPixel(x + sX, y + sX, color.rgba());
		image.setPixel(x + sX, y - sX, color.rgba());
	}
}

void drawCircle(int x, int y, int sideSize, const QColor &color, QImage &image)
{
	int numPixels = (qCeil(2 * PI * sideSize/2) + 1) * 2;

	for(int i = 0; i < numPixels; i++)
	{
		float alpha = (float) i / (float) sideSize;
		alpha *= 2.0f * PI;
		int sX = qFloor(cos(alpha) * sideSize/2);
		int sY = qFloor(sin(alpha) * sideSize/2);

		image.setPixel(x + sX, y + sY, color.rgba());
	}
}

float colorDist(feature_t *F1, feature_t *F2)
{
	int dR = F1->r - F2->r, dG = F1->g - F2->g, dB = F1->b - F2->b;
	return qSqrt(dR*dR + dG*dG + dB*dB); 
}

float luminance(const QColor &col)
{
	return 0.2125f * col.redF() + 0.7154 * col.greenF() + 0.0721 * col.blueF();
}

float erf(float x)
{
	// constants
	float a1 =  0.254829592f;
	float a2 = -0.284496736f;
	float a3 =  1.421413741f;
	float a4 = -1.453152027f;
	float a5 =  1.061405429f;
	float p  =  0.3275911f;

	// Save the sign of x
	int sign = 1;
	if (x < 0)
		sign = -1;
	x = fabs(x);

	// A&S formula 7.1.26
	float t = 1.0f/(1.0f + p*x);
	float y = 1.0f - (((((a5*t + a4)*t) + a3)*t + a2)*t + a1)*t*exp(-x*x);

	return sign*y;
}