#ifndef TOOLS_H
#define TOOLS_H

#include <QVector2D>
#include <QVector3D>
#include <QtCore/qmath.h>
#include <QImage>
#include <QColor>

extern "C" {
#include "emd.h"
}

using namespace std;

/**
 * Stores pixels of an image in an QVector3D.
 * @param x X coordinate.
 * @param y Y coordinate.
 * @param width Width of the part that should be tiled.
 * @param height Height of the part that should be tiled.
 * @param orig Image used to create the vector of pixels.
 * @return Vector of pixels of the part of the image.
 */
vector<QVector3D> createSubImage(
	const int &x, const int &y,
	const int &width, const int &height,
	const QImage &orig);

/**
 * Creates a QImage using the x, y width and height of a QRect.
 * @param image Image used.
 * @param rect Rectangle used to get coordinates and size.
 * @return Sub image.
 */
QImage createSubImage(
	const QImage &image, const QRect & rect);

/**
 * Tiles a part of an image.
 * Uses a part of an image to tile it into two sets of pixel values. A gradient is used to separate the part of the image ("left and right" from the vector).
 * @param x X coordinate.
 * @param y Y coordinate.
 * @param width Width of the part that should be tiled.
 * @param height Height of the part that should be tiled.
 * @param gradient 2D gradient vector used to tile the part.
 * @param image Image used for color information.
 * @param subImage1 Vector of pixels for the one half of the part of the image.
 * @param subImage2 Vector of pixels for the second half of the part of the image.
 */
void createSubImage(
	const int &x, const int &y,
	const int &width, const int &height,
	const QVector2D &gradient, const QImage &orig,
	vector<QVector3D> &subImage1, vector<QVector3D> &subImage2);

/**
 * Draws a sub image into an image with given coordinates.
 * @param x X coordinate.
 * @param y Y coordinate.
 * @param subImage Sub image which should be drawn into the image.
 * @param image The image into which the sub image is drawn.
 */
void drawSubImage(int x, int y, const QImage &subImage, QImage &image);
/**
 * Draws a square into an image with given color, coordinates and side width.
 * @param x X coordinate.
 * @param y Y coordinate.
 * @param sideSize Width of a square's side.
 * @param color Color used to draw the square.
 * @param image The image into which the square is drawn.
 */
void drawSubImage(int x, int y, int sideSize, const QColor &color, QImage &image);
/**
 * Draws a cross into an image with given color, coordinates and side width.
 * @param x X coordinate.
 * @param y Y coordinate.
 * @param sideSize Width of a square's side.
 * @param color Color used to draw the cross.
 * @param image The image into which the cross is drawn.
 */
void drawCross(int x, int y, int sideSize, const QColor &color, QImage &image);
/**
 * Draws a circle into an image with given color, coordinates and diagonal size.
 * @param x X coordinate.
 * @param y Y coordinate.
 * @param sideSize Diagonal of the circle.
 * @param color Color used to draw the circle.
 * @param image The image into which the circle is drawn.
 */
void drawCircle(int x, int y, int sideSize, const QColor &color, QImage &image);

/**
 * Calculates the distances between two color features.
 * @param F1 First color signature feature.
 * @param F2 Second color signature feature.
 */
float colorDist(feature_t *F1, feature_t *F2);

/**
 * Calculates the luminance of a given color.
 * @param color Color used to calculate.
 * @return Luminance value.
 */
float luminance(const QColor &color);

/**
 * Mirrors an image horizontally.
 * @param image Original image.
 * @return Mirrored image.
 */
QImage flipImageHorizontally(const QImage &image);

/**
 * Computes the error function.
 */
float erf(float x);

#endif