#ifndef TRANSFERPOINT_HPP_
#define TRANSFERPOINT_HPP_

#include <QGraphicsItem>
#include <QGraphicsSceneMouseEvent>

/**
 * Class which represents a point at the transfer function.
 * Holds a color, alpha value and the position.
 * @author Christian Hirsch
 * @author Clemens Roegner
 */
class TransferPoint : public QGraphicsItem
{
public:
	/**
	 * Constructor.
	 * Creates a new point at the transfer function at the relative x and y position.
	 * @param relX The relative x position. If position is 1.0 or 0.0 then the point can't be moved along the x axis.
	 * @param relY The relative y position. Represents the alpha value of a point.
	 */
	TransferPoint(qreal relX, qreal relY, qreal sigma = 0.2);

	/**
	 * Returns the bounding rectangle of the point.
	 * @return The bounding rectangle of the point.
	 */
	QRectF boundingRect() const;

	/**
	 * Redraws the scene of the transfer widget.
	 */
	void updateScene();
	/**
	 * Paints the point with a black bounding box, filled with its color, which is represented.
	 */
	void paint (QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget);

	/**
	 * Sets if a point is fixed (can't be moved along x axis) or not.
	 * @param fixed If true, then the point is fixed.
	 */
	void setFixed(bool fixed);
	/**
	 * Setter for the relative x and y Position, where x = 1.0 is the right side, and y = 1.0 is at the top.
	 * @param x New relative x position.
	 * @param y New relative y position.
	 */
	void setRelativePos(qreal x, qreal y);
	/**
	 * Getter for relative x position.
	 * @return Relative x position.
	 */
	qreal getRelativeX();
	/**
	 * Getter for relative y position.
	 * @return Relative y position.
	 */
	qreal getRelativeY();
	/**
	 * Getter for sigma value.
	 * @return Sigma value of control point.
	 */
	qreal getSigma();
	/**
	 * Setter for sigma value.
	 * @param sigma Sigma value of control point.
	 */
	void setSigma(qreal sigma);

	/**
	 * Sets the representing color of the point.
	 * @param _color New representing color.
	 */
	void setColor(QColor _color);
	/**
	 * Opens a color-dialog, in which the user can decide which color to use.
	 */
	void setColorFromDialog();
	/**
	 * Getter for the color.
	 * return The representing color.
	 */
	QColor getColor();

protected:
	/**
	 * Handles the mouse press event. (Nothing is done, but mousemove event needs this).
	 * @param event Pointer to the called event.
	 */
	void mousePressEvent(QGraphicsSceneMouseEvent *event);
	/**
	 * Handles the mouse move event. Moves a point on the transfer widget.
	 */
	void mouseMoveEvent(QGraphicsSceneMouseEvent *event);

private:
	qreal relX;
	qreal relY;
	qreal sigma;
	bool isFixed;
	QColor color;

	/**
	 * Updates the position with the new x and y coordinates.
	 * @param x New x position.
	 * @param y New y position.
	 */
	void updatePosition(qreal x, qreal y);
};

#endif /* TRANSFERPOINT_HPP_ */
