#include <QtGui>
#include <QGraphicsRectItem>
#include <QMouseEvent>
#include <stdio.h>

#include "TransferWidget.hpp"
#include "TransferPoint.hpp"

const float PI = acos(-1.0f);

TransferWidget::TransferWidget(QWidget *parent = 0)
: QGraphicsView(parent)
{
	// turn off scroll bars
	setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
	setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);

	// scene initialization
	scene = new QGraphicsScene(this);
	setScene(scene);

	TransferPoint *pnt = new TransferPoint(0, 1);
	pnt->setFixed(true);
	pnt->setColor(Qt::black);
	scene->addItem(pnt);
	points.append(pnt);

	pnt = new TransferPoint(1, 0);
	pnt->setFixed(true);
	pnt->setColor(Qt::white);
	scene->addItem(pnt);
	points.append(pnt);

	QGraphicsLineItem *line = new QGraphicsLineItem(
			points[0]->x(), points[0]->y(),
			points[1]->x(), points[1]->y());
	lines.append(line);
	scene->addItem(line);

	setRenderHint(QPainter::Antialiasing);
	setBackgroundBrush(QPixmap("res/background.png"));
	setCacheMode(QGraphicsView::CacheBackground);

	fillBackground();
}

void TransferWidget::mousePressEvent(QMouseEvent *event)
{
	QGraphicsView::mousePressEvent(event);

	if(event->button() == Qt::RightButton)
	{
		TransferPoint *pnt = new TransferPoint(
			event->pos().x() / scene->width(),
			event->pos().y() / scene->height());
		pnt->setColorFromDialog();
		scene->addItem(pnt);

		// put to correct position
		int i;
		for(i=0; points[i]->getRelativeX() < pnt->getRelativeX(); i++)
			;
		points.insert(i, pnt);
	}
	scene->removeItem(background);
	redrawScene();
}

void TransferWidget::mouseMoveEvent(QMouseEvent *event)
{
	QGraphicsView::mouseMoveEvent(event);

	redrawScene();
}

void TransferWidget::resizeEvent(QResizeEvent *event)
{
	scene->setSceneRect(0, 0, width()-2, height()-2);
	redrawScene();
}

void TransferWidget::redrawScene()
{
	points[0]->updateScene();
	// update positions of transfer points
	for(int i=1; i<points.size(); i++)
	{
		points[i]->updateScene();

		// swap points
		if(points[i-1]->x() > points[i]->x())
			points.swap(i-1, i);
	}
	scene->removeItem(background);
	fillBackground();
}

void TransferWidget::fillBackground()
{
	QLinearGradient gradient(scene->sceneRect().bottomLeft(), scene->sceneRect().bottomRight());
	QPainterPath *bg = new QPainterPath(scene->sceneRect().bottomLeft());

	for(int i=0; i<points.size(); i++)
	{
		gradient.setColorAt(points[i]->getRelativeX(), points[i]->getColor());
		bg->lineTo(points[i]->pos());
	}
	bg->lineTo(scene->sceneRect().bottomRight());
	bg->lineTo(scene->sceneRect().bottomLeft());

	background = scene->addPath(*bg, QPen(Qt::black, 1), QBrush(gradient));

	emit transferChanged();
}

std::vector<TransferWidget::TransferFeature> TransferWidget::getTransferControlValues()
{
	std::vector<TransferFeature> transferFeatures;

	for(int i=0; i<points.size(); i++)
	{
		TransferFeature feature;
		feature.pos = (float) points[i]->getRelativeX();
		feature.alpha = 1.0f - (float) points[i]->getRelativeY();
		feature.color = points[i]->getColor();
		feature.sigma = (float) points[i]->getSigma();

		transferFeatures.push_back(feature);
	}
	return transferFeatures;
}

int TransferWidget::getNumPoints()
{
	return points.size();
}

void TransferWidget::clear()
{
	points.clear();

	scene->clear();

	// scene initialization
	scene = new QGraphicsScene(this);
	setScene(scene);

	fillBackground();
}

void TransferWidget::setNewRGBXY(float _red, float _green, float _blue, float _x, float _y)
{
	QColor *color = new QColor;
	
	color->setRedF(_red);
	color->setGreenF(_green);
	color->setBlueF(_blue);
	color->setAlphaF(1.0);

	TransferPoint *pnt = new TransferPoint(_x, _y);
	pnt->setColor(*color);
	scene->addItem(pnt);
	points.append(pnt);

	//redrawScene();
}