#ifndef TRANSFORMWIDGET_HPP_
#define TRANSFORMWIDGET_HPP_

#include <QGraphicsView>
#include <QGraphicsScene>
#include <QPainterPath>
#include <QList>

#include "TransferPoint.hpp"

/**
 * The transfer widget in which the transfer funtion can be set.
 * @author Christian Hirsch
 * @author Clemens Roegner
 */
class TransferWidget : public QGraphicsView
{
	Q_OBJECT

public:

	/**
	 * Structure that holdes features of the control point.
	 */
	struct TransferFeature {
		float pos;
		float sigma;
		float alpha;
		QColor color;
	};

	/**
	 * Constructor.
	 * Creates a new class, which represents the transfer function.
	 * @param parent Pointer to the parent QWidget element.
	 */
	TransferWidget(QWidget *parent);

	/**
	 * Redraws the Scene.
	 * Redraws the scene with the current set transfer function. Under the points you can see the color-flow, which will be applied to the volume data.
	 */
	void redrawScene();

	/**
	 * Returns a float pointer to the transfer function control values, which are used to store in an file. (Format: RGBXY)
	 * @return Pointer to float values of the control points.
	 */
	std::vector<TransferFeature> getTransferControlValues();

	/**
	 * Retuns the number of current points.
	 * @return Number of current points.
	 */
	int getNumPoints();

	/**
	 * Removes the current transfer function values.
	 */
	void clear();

	/**
	 * Creates a new control point for the transfer function.
	 * @param _red The red value in range 0.0 - 1.0
	 * @param _green The green value in range 0.0 - 1.0
	 * @param _blue The blue value in range 0.0 - 1.0
	 * @param _x The relative x position of the new control point in range 0.0 - 1.0
	 * @param _y The relative y position of the new control point in range 0.0 - 1.0
	 */
	void setNewRGBXY(float _red, float _green, float _blue, float _x, float _y);

	/**
	 * Handles the resize event.
	 * When the transfer widget resizes, all control point positions should be updated and the transfer widget should be redrawn.
	 * @param event Pointer to the calling event.
	 */
	void resizeEvent(QResizeEvent *event);

signals:
	void transferChanged();

private:
	QGraphicsScene *scene;
	QGraphicsPathItem *background;
	QList<QGraphicsLineItem *> lines;
	QList<TransferPoint *> points;

	/**
	 * Fills the background under the control points in the color, which will be applyed to the volume data.
	 */
	void fillBackground();
	
	/**
	 * Handles the mouse press event.
	 * When the right button is pressed a new control point at the current position will be added.
	 * @param event Pointer to the calling event.
	 */
	void mousePressEvent(QMouseEvent *event);

	/**
	 * Handles the mouse move event.
	 * When the mouse moves on the transfer widget, it will be redrawn while moving.
	 * @param event Pointer to the calling event.
	 */
	void mouseMoveEvent(QMouseEvent *event);
};

#endif /* TRANSFORMWIDGET_HPP_ */
