
#include <QtCore/qmath.h>

#include "cannyedge.h";

CannyEdge::CannyEdge(QWindow *_qgl_window, QOpenGLContext *_qgl_context, float _sigma, int _filterSize, int _width, int _height)
{
	sigma = _sigma;
	filterSize = _filterSize;
	width = _width;
	height = _height;

	qgl_window = _qgl_window;
	qgl_context = _qgl_context;
	
	// Obtain a functions object and resolve all entry points
    // m_funcs is declared as: QOpenGLFunctions_3_2_Core* v_funcs
    qgl_funcs = qgl_context->versionFunctions<QOpenGLFunctions_3_2_Core>();
    if ( !qgl_funcs ) {
        qCritical() << "Could not obtain OpenGL versions object";
        exit( 1 );
    }
    qgl_funcs->initializeOpenGLFunctions();
	
	qgl_context->makeCurrent(qgl_window);

	qgl_funcs->glGenTextures(1, &filterTex);

	QOpenGLFramebufferObjectFormat formatFBO;
	formatFBO.setAttachment(QOpenGLFramebufferObject::NoAttachment);
	formatFBO.setMipmap(false);
	formatFBO.setTextureTarget(GL_TEXTURE_RECTANGLE);
	formatFBO.setInternalTextureFormat(GL_RGBA16);
	qgl_cannyFBO1 = new QOpenGLFramebufferObject(
		width, height, formatFBO);
	if(!qgl_cannyFBO1->isValid())
		qDebug() << "Failed to create QOpenGLFramebufferObject \'qgl_cannyFBO1\'";
	qgl_cannyFBO2 = new QOpenGLFramebufferObject(
		width, height, formatFBO);
	if(!qgl_cannyFBO2->isValid())
		qDebug() << "Failed to create QOpenGLFramebufferObject \'qgl_cannyFBO2\'";

	prepareShaders();
	prepareVertexArrayObject();
	filterCoefficients = new float[0];
	updateFilterCoefficients();
}

CannyEdge::~CannyEdge()
{
	delete[] filterCoefficients;
	//delete[] verticesFullscreenQuad;
	delete(qgl_cannyFBO1);
	delete(qgl_cannyFBO2);
}

void CannyEdge::setSigma(float _sigma)
{
	sigma = _sigma;
}

float CannyEdge::getSigma()
{
	return sigma;
}

void CannyEdge::setFilterSize(int _filterSize)
{
	filterSize = _filterSize;
}

int CannyEdge::getFilterSize()
{
	return filterSize;
}

void CannyEdge::updateFilterCoefficients()
{
	delete[] filterCoefficients;

	filterCoefficients = new float[filterSize];
	for(int i=0; i < filterSize; i++)
	{
		filterCoefficients[i] = 1.0f / (qSqrt(2.0f * PI) * sigma)
			* qExp(-1.0f*(qPow((i - qFloor(filterSize/2)), 2.0f) / (2 * qPow(sigma, 2))));
	}

	qgl_funcs->glBindTexture(GL_TEXTURE_1D, filterTex);
	qgl_funcs->glTexImage1D(GL_TEXTURE_1D, 0,
		GL_RED, filterSize, 0, GL_RED, GL_FLOAT, filterCoefficients);
	qgl_funcs->glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	qgl_funcs->glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
}

GLuint CannyEdge::getGradientFieldTexture()
{
	return qgl_cannyFBO2->texture();
}

GLuint CannyEdge::getCannyEdgeTexture()
{
	return qgl_cannyFBO1->texture();
}

QImage CannyEdge::getGradientFieldImage()
{
	return qgl_cannyFBO2->toImage();
}

QImage CannyEdge::getCannyEdgeImage()
{
	return qgl_cannyFBO1->toImage();
}

void CannyEdge::detectEdges(GLuint texture)
{
	imageBlur(qgl_cannyFBO1, texture);
	imageMagnitude(qgl_cannyFBO2, qgl_cannyFBO1->texture());
	imageNonMaximaSuppression(qgl_cannyFBO1, qgl_cannyFBO2->texture());
}

void CannyEdge::imageBlur(QOpenGLFramebufferObject *fbo, GLuint texture)
{
	// BLUR IMAGE
	fbo->bind();

	qgl_funcs->glDisable(GL_CULL_FACE);
	qgl_funcs->glDisable(GL_DEPTH_TEST);

	qgl_funcs->glActiveTexture(GL_TEXTURE0);
	qgl_funcs->glBindTexture(GL_TEXTURE_1D, filterTex);
	qgl_funcs->glActiveTexture(GL_TEXTURE1);
	qgl_funcs->glBindTexture(GL_TEXTURE_RECTANGLE, texture);

	qgl_cannyBlurShader.bind();
	qgl_cannyBlurShader.setUniformValue("width", (float)width);
	qgl_cannyBlurShader.setUniformValue("height", (float)height);
	qgl_cannyBlurShader.setUniformValue("filterSize", filterSize);
	qgl_cannyBlurShader.setUniformValue("horizontal", true);
	qgl_cannyBlurShader.setUniformValue("filterCoefficients", GLint(0));
	qgl_cannyBlurShader.setUniformValue("image", GLint(1));

	qgl_funcs->glClear(GL_COLOR_BUFFER_BIT);
	QOpenGLVertexArrayObject::Binder binder( &qgl_quad );
	qgl_funcs->glDrawArrays(GL_TRIANGLES, 0, 6);

	qgl_funcs->glActiveTexture(GL_TEXTURE1);
	qgl_funcs->glBindTexture(GL_TEXTURE_RECTANGLE, fbo->texture());
	qgl_cannyBlurShader.setUniformValue("horizontal", false);

	qgl_funcs->glDrawArrays(GL_TRIANGLES, 0, 6);

	qgl_cannyBlurShader.release();
	fbo->release();
}

void CannyEdge::imageMagnitude(QOpenGLFramebufferObject *fbo, GLuint texture)
{
	// DETERMINE MAGNITUDE
	fbo->bind();
	qgl_funcs->glDisable(GL_CULL_FACE);
	qgl_funcs->glDisable(GL_DEPTH_TEST);

	qgl_funcs->glActiveTexture(GL_TEXTURE0);
	qgl_funcs->glBindTexture(GL_TEXTURE_RECTANGLE, texture);

	qgl_cannyMagnitudeShader.bind();
	qgl_cannyMagnitudeShader.setUniformValue("width", (float)width);
	qgl_cannyMagnitudeShader.setUniformValue("height", (float)height);
	qgl_cannyMagnitudeShader.setUniformValue("image", GLint(0));

	qgl_funcs->glClear(GL_COLOR_BUFFER_BIT);
	QOpenGLVertexArrayObject::Binder binder( &qgl_quad );
	qgl_funcs->glDrawArrays(GL_TRIANGLES, 0, 6);

	qgl_cannyMagnitudeShader.release();
	fbo->release();
}

void CannyEdge::imageNonMaximaSuppression(QOpenGLFramebufferObject *fbo, GLuint texture)
{
	// NON MAXIMA SUPPRESSION
	fbo->bind();
	qgl_funcs->glDisable(GL_CULL_FACE);
	qgl_funcs->glDisable(GL_DEPTH_TEST);

	qgl_funcs->glActiveTexture(GL_TEXTURE0);
	qgl_funcs->glBindTexture(GL_TEXTURE_RECTANGLE, texture);

	qgl_cannyEdgeShader.bind();
	qgl_cannyEdgeShader.setUniformValue("width", (float)width);
	qgl_cannyEdgeShader.setUniformValue("height", (float)height);
	qgl_cannyEdgeShader.setUniformValue("image", GLint(0));
	qgl_cannyEdgeShader.setUniformValue("thresh", 0.01f);

	qgl_funcs->glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	QOpenGLVertexArrayObject::Binder binder( &qgl_quad );
	qgl_funcs->glDrawArrays(GL_TRIANGLES, 0, 6);

	qgl_cannyEdgeShader.release();

	fbo->release();
}

void CannyEdge::prepareShaders()
{
	if(!qgl_cannyBlurShader.addShaderFromSourceFile(QOpenGLShader::Vertex, "res/CannyShader.vert"))
		qCritical() << "Could not compile vertex shader. Log: " << qgl_cannyBlurShader.log();
	if(!qgl_cannyBlurShader.addShaderFromSourceFile(QOpenGLShader::Fragment, "res/CannyBlurShader.frag"))
		qCritical() << "Could not compile fragment shader. Log: " << qgl_cannyBlurShader.log();
	qgl_cannyBlurShader.link();

	if(!qgl_cannyMagnitudeShader.addShaderFromSourceFile(QOpenGLShader::Vertex, "res/CannyShader.vert"))
		qCritical() << "Could not compile vertex shader. Log: " << qgl_cannyMagnitudeShader.log();
	if(!qgl_cannyMagnitudeShader.addShaderFromSourceFile(QOpenGLShader::Fragment, "res/CannyMagnitudeShader.frag"))
		qCritical() << "Could not compile fragment shader. Log: " << qgl_cannyMagnitudeShader.log();
	qgl_cannyMagnitudeShader.link();

	if(!qgl_cannyEdgeShader.addShaderFromSourceFile(QOpenGLShader::Vertex, "res/CannyShader.vert"))
		qCritical() << "Could not compile vertex shader. Log: " << qgl_cannyEdgeShader.log();
	if(!qgl_cannyEdgeShader.addShaderFromSourceFile(QOpenGLShader::Fragment, "res/CannyEdgeShader.frag"))
		qCritical() << "Could not compile fragment shader. Log: " << qgl_cannyEdgeShader.log();
	qgl_cannyEdgeShader.link();
}

void CannyEdge::prepareVertexArrayObject()
{
	const int vertexCount = 6;

	// Setup VBOs and IBO (use QOpenGLBuffer to buffer data,
    // specify format, usage hint etc). These will be
    // remembered by the currently bound VAO
	float vertexPositions[] = {
		-1.0f,  1.0f,  0.0f,
		-1.0f, -1.0f,  0.0f,
		 1.0f, -1.0f,  0.0f,
		 1.0f, -1.0f,  0.0f,
		 1.0f,  1.0f,  0.0f,
		-1.0f,  1.0f,  0.0f
	};
	QOpenGLBuffer qgl_positionBuffer;
    qgl_positionBuffer.create();
	qgl_positionBuffer.setUsagePattern( QOpenGLBuffer::StaticDraw );
    qgl_positionBuffer.bind();
	qgl_positionBuffer.allocate(
		vertexPositions,
        vertexCount * 3 * sizeof(float)
		);
	qgl_positionBuffer.release();

	qgl_quad.create();
	QOpenGLVertexArrayObject::Binder binder( &qgl_quad );

	binder.rebind();
	qgl_positionBuffer.bind();
	qgl_cannyBlurShader.enableAttributeArray("inputVertexPosition");
	qgl_cannyBlurShader.setAttributeBuffer("inputVertexPosition", GL_FLOAT, 0, 3);
	qgl_cannyMagnitudeShader.enableAttributeArray("inputVertexPosition");
	qgl_cannyMagnitudeShader.setAttributeBuffer("inputVertexPosition", GL_FLOAT, 0, 3);
	qgl_cannyEdgeShader.enableAttributeArray("inputVertexPosition");
	qgl_cannyEdgeShader.setAttributeBuffer("inputVertexPosition", GL_FLOAT, 0, 3);
	qgl_positionBuffer.release();

	binder.release();
}