#ifndef CANNYEDGE_H
#define CANNYEDGE_H

#include <QtGui>
#include <QWindow>
#include <QtOpenGL>
#include <QOpenGLFunctions_3_2_Core>
#include "junctiondetector.h"

#define PI acos(-1.0f)


/** 
 *	Class to detect canny edge using OpenGL.
 *	The CannyEdge class stores data to create an image containing red edges that are created using the canny edge detector.
 *	@author Clemens Roegner
 *  @author Christian Hirsch
 */
class CannyEdge
{
public:
	/**
	 * Constructor.
	 * Creates a new CannyEdge object with all necessary settings.
	 * @param _qgl_window Pointer to the Window where rendering is done.
	 * @param _qgl_contex Pointer to the used and initialized OpenGL context.
	 * @param _sigma Sigma used for gaussian blurring of the image.
	 * @param _filterSize Filterkernel size used for the gaussion blurring of the image.
	 * @param _width Window width.
	 * @param _height Window height.
	 */ 
	CannyEdge(QWindow *_qgl_window, QOpenGLContext *_qgl_contex, float _sigma, int _filterSize, int _width, int _height);

	/**
	 * Destructor.
	 */
	~CannyEdge();

	/**
	 * Setter for the sigma value.
	 * @param _sigma New value of sigma used for the guassian blurring of the image.
	 */
	void setSigma(float _sigma);
	/**
	 * Getter for sigma value.
	 * @return The sigma value used for the gaussian blurring of the image.
	 */
	float getSigma();
	/**
	 * Setter for the kernel filtersize value.
	 * @param _filterSize New value of the kernel filtersize used for the guassian blurring of the image.
	 */
	void setFilterSize(int _filterSize);
	/**
	 * Getter for kernelfilter size value.
	 * @return The kernel filtersize value used for the gaussian blurring of the image.
	 */
	int getFilterSize();

	/**
	 * Detects canny edges on an given texture.
	 */
	void detectEdges(GLuint texture);

	/**
	 * Getter for the gradient field texture handler.
	 * @return The texture handler in which the gradient fields are stored.
	 */
	GLuint getGradientFieldTexture();
	/**
	 * Getter for the canny edge texture handler.
	 * @return The texture handler in which the canny edges are stored.
	 */
	GLuint getCannyEdgeTexture();
	/**
	 * Getter for the gradient field image.
	 * @return A new QImage in which the gradient fields are stored.
	 */
	QImage getGradientFieldImage();
	/**
	 * Getter for the gradient field image.
	 * @return A new QImage in which the gradient fields are stored.
	 */
	QImage getCannyEdgeImage();

private:
	float sigma;
	int filterSize;
	float *filterCoefficients;
	int width, height;
	float *verticesFullscreenQuad;

	QWindow *qgl_window;

	GLuint image;
	GLuint filterTex;

	QOpenGLContext *qgl_context;
	QOpenGLFunctions_3_2_Core *qgl_funcs;
	
	QOpenGLShaderProgram qgl_cannyBlurShader;
	QOpenGLShaderProgram qgl_cannyMagnitudeShader;
	QOpenGLShaderProgram qgl_cannyEdgeShader;
	QOpenGLFramebufferObject *qgl_cannyFBO1;
	QOpenGLFramebufferObject *qgl_cannyFBO2;

	QOpenGLVertexArrayObject qgl_quad;

	void prepareShaders();
	void prepareVertexArrayObject();

	void updateFilterCoefficients();

	void imageBlur(QOpenGLFramebufferObject *fbo, GLuint texture);
	void imageMagnitude(QOpenGLFramebufferObject *fbo, GLuint texture);
	void imageNonMaximaSuppression(QOpenGLFramebufferObject *fbo, GLuint texture);
};

#endif //CANNYEDGE