#include "junctiondetector.h"
#include <QColor>
#include <QtGui>
#include "Tools.h"

#define assert(isTrue) if(!(isTrue))
//qWarning() << "Assertion triggered!"

const float PI = acos(-1.0f);

unsigned int JunctionDetector::recursiveDepth = 0;
const QVector2D JunctionDetector::normalizedVector = QVector2D(
	1.0f / qSqrt(2.0f), 1.0f / qSqrt(2.0f));

void JunctionDetector::detectJunctions(const QImage &image,
	const QImage &cannyEdgeImage,
	const QImage &gradientImage,
	QImage &junctionImage,
	QImage &outputJunctionImage)
{
	outputJunctionImage.fill(QColor(0, 0, 0, 0));
	QVector2D normalizedVector(1.0f, 1.0f);
	normalizedVector.normalize();
	float maxEmdValue = 0.0f;
	float overallMeanEmd = 0.0f;
	float overallSigmaEmd = 0.0f;
	vector<QVector3D> emdValues;

	for(int x=0; x < image.width(); x++)
	{
		for(int y=0; y < image.height(); y++)
		{
			QRgb cannyRgb = cannyEdgeImage.pixel(x, y);
			if(qRed(cannyRgb) != 255) {
				junctionImage.setPixel(x, junctionImage.height()-y-1, qRgba(0, 0, 0, 0));
				continue;
			}

			QRgb gradientRgb = gradientImage.pixel(x, y);
			QVector2D gradientVector(
				qRed(gradientRgb), qGreen(gradientRgb));

			gradientVector = gradientVector * (2.0f / 255.0f) - normalizedVector;

			vector<QVector3D> subPixels1;
			vector<QVector3D> subPixels2;
			createSubImage(x, y,
				5, 5, gradientVector, image,
				subPixels1, subPixels2);
			if(subPixels1.size() == 0
				|| subPixels2.size() ==0)
				continue;

			signature_t colorSignature1, colorSignature2;
			JunctionDetector::calculateColorsignature(subPixels1, colorSignature1);
			JunctionDetector::calculateColorsignature(subPixels2, colorSignature2);

			// calculate emd
			float emdValue = emd(&colorSignature1, &colorSignature2, colorDist, 0, 0);
			overallMeanEmd += emdValue;
			if(emdValue > maxEmdValue)
				maxEmdValue = emdValue;
			emdValues.push_back(QVector3D(x, junctionImage.height()-y-1, emdValue));

			free(colorSignature1.Features); free(colorSignature1.Weights);
			free(colorSignature2.Features); free(colorSignature2.Weights);
		}
	}

	overallMeanEmd /= emdValues.size();

	qDebug() << "EMD values calculated";

	for(int i=0; i < emdValues.size(); i++)
	{
		overallSigmaEmd += qPow(emdValues[i].z() - overallMeanEmd, 2);
		junctionImage.setPixel(emdValues[i].x(), emdValues[i].y(),
			qRgba(emdValues[i].z() / maxEmdValue * 255.0f, 0, 0, 255));
	}

	overallMeanEmd = overallMeanEmd / maxEmdValue * 255.0f;
	overallSigmaEmd = overallSigmaEmd / maxEmdValue * 255.0f;

	float thresh = overallMeanEmd + (255.0f - overallMeanEmd) / 4.0f * 2.0f;
	thresh = 0.9f;

	for(int x = 0; x < junctionImage.width(); x++)
	{
		for(int y = 0; y < junctionImage.height(); y++)
		{
			float emd = qRed(junctionImage.pixel(x, y));
			float cdf = 0.5f * (1.0f + erf((emd - overallMeanEmd) / (overallSigmaEmd * (2.0f))));
			if(emd < thresh)
				continue;

			float meanEmd = 0.0f;
			int N = 0;
			bool biggest = true;
			const int windowSize = 15;
			for(int subX = x-windowSize/2; subX < x+windowSize/2+1; subX++)
			{
				for(int subY = y-windowSize/2; subY < y+windowSize/2+1; subY++)
				{
					float emdSub = qRed(junctionImage.pixel(subX, subY));
					if(!(emdSub > 0))
						continue;
					if(emdSub > emd)
						biggest = false;
					if(!biggest)
						continue;

					N++;
					meanEmd += qRed(junctionImage.pixel(subX, subY));
				}
			}

			if(!(N>0) || !biggest)
				continue;

			meanEmd /= (float)N;
			if(emd > (meanEmd + (255.0f - meanEmd) / 10.0f * 2.0f))
			{
				junctionImage.setPixel(x, y, qRgba(255, 255, 0, 255));
				drawCircle(x, y, 15, QColor(0, 0, 130, 255), outputJunctionImage);
			}
		}
	}
}

void JunctionDetector::detectXJunctions(
	QImage &image,
	QImage &junctionImage,
	QImage &outputXJunctionImage,
	std::vector<std::vector<float>> &luminances)
{
	outputXJunctionImage.fill(QColor(0, 0, 0, 0));

	int imgWidth = junctionImage.width();
	int imgHeight = junctionImage.height();

	const int windowSize = 15;

	QColor colors[5] = {
		QColor(0, 0, 0, 255),
		QColor(255, 255, 0, 255),	// yellow
		QColor(0, 255, 255, 255),	// cyan
		QColor(0, 255, 0, 255),		// green
		QColor(0, 0, 255, 255)		// blue
	};

	for(int x=windowSize; x < imgWidth - windowSize; x++)
	{
		for(int y=windowSize; y < imgHeight - windowSize; y++)
		{
			if(qGreen(junctionImage.pixel(x, y)) == 0)
				continue;

			QImage neightbours(windowSize, windowSize, QImage::Format_Indexed8);
			QImage neighbourJunctions = createSubImage(image,
				QRect(x - windowSize / 2, y - windowSize / 2, windowSize, windowSize));
			
			int isX = 0;
			for(float startAlpha = -PI/8.0f; startAlpha < 3 * PI/8.0f; startAlpha += PI/8.0f)
			{
				neightbours.setColorCount(1);
				neightbours.fill(0);
				for(float alpha = startAlpha; alpha < 2.0f * PI; alpha += PI / 2.0f)
				{
					int seedX = qFloor((windowSize - 1) / 2.0f * (1 + cos(alpha)));
					int seedY = qFloor((windowSize - 1) / 2.0f * (1 + sin(alpha)));

					unsigned int value = neightbours.pixel(seedX, seedY);
					if(value != 0)
						break;
					
					QColor initColor(neighbourJunctions.pixel(seedX, seedY));
					neightbours.setColor(neightbours.colorCount(), initColor.rgb());

					growRegion(seedX, seedY, neighbourJunctions, initColor, neightbours);
				}
				if(neightbours.colorCount() > 4)
					isX ++;
			}
			if(isX > 1)
			{
				std::vector<float> lums;
				lums.push_back(luminance(QColor(neighbourJunctions.pixel(0, 0))));
				lums.push_back(luminance(QColor(neighbourJunctions.pixel(0, neighbourJunctions.height()-1))));
				lums.push_back(luminance(QColor(neighbourJunctions.pixel(neighbourJunctions.width()-1, 0))));
				lums.push_back(luminance(QColor(neighbourJunctions.pixel(neighbourJunctions.width()-1, neighbourJunctions.height()-1))));
				
				for(int j = 0; j < lums.size()-1; j++)
					for(int i=1; i < lums.size(); i++)
					{
						if(lums[i] > lums[i-1])
						{
							float lum = lums[i];
							lums[i] = lums[i-1];
							lums[i-1] = lum;
						}
					}
				luminances.push_back(lums);
				drawCross(x, y, 15, QColor(0, 130, 0, 255), outputXJunctionImage);
			}
		}
	}
}

void JunctionDetector::growRegion(
	const int &seedX, const int &seedY,
	QImage &image,
	const QColor &initialColor,
	QImage &neighbours)
{
	if(seedX >= neighbours.width())
		return;
	if(seedY >= neighbours.height())
		return;
	if(seedX < 0)
		return;
	if(seedY < 0)
		return;
	if(neighbours.pixel(seedX, seedY) > 0)
		return;

	QColor pixelColor(image.pixel(seedX, seedY));

	if(colorDiff(initialColor, pixelColor) > 0.15f)
		return;
	neighbours.setPixel(seedX, seedY, neighbours.colorCount() - 1);

	growRegion(seedX + 1, seedY, image, initialColor, neighbours);
	growRegion(seedX, seedY + 1, image, initialColor, neighbours);
	growRegion(seedX - 1, seedY, image, initialColor, neighbours);
	growRegion(seedX, seedY - 1, image, initialColor, neighbours);
}

float JunctionDetector::colorDiff(
	const QColor &color1, const QColor &color2)
{
	float red = qAbs(color1.redF() - color2.redF());
	float green = qAbs(color1.greenF() - color2.greenF());
	float blue = qAbs(color1.blueF() - color2.blueF());
	float alpha = qAbs(color1.alpha() - color2.alpha());

	//return qSqrt(red * red + green * green + blue * blue);
	return qSqrt(red * red + green * green + blue * blue + alpha * alpha);
}

QMatrix3x3 JunctionDetector::multiplyVectors(
	const QVector3D *vec,
	const QVector3D *vecT)
{
	float values[9] = {
		vec->x() * vecT->x(), vec->x() * vecT->y(), vec->x() * vecT->z(),
		vec->y() * vecT->x(), vec->y() * vecT->y(), vec->y() * vecT->z(),
		vec->z() * vecT->x(), vec->z() * vecT->y(), vec->z() * vecT->z()
	};

	return QMatrix3x3(values);
}

void JunctionDetector::calculateColorsignature(
	const vector<QVector3D> &imagePixels,
	signature_t &signature)
{
	vector<ColorSignature> colorSignatures;

	// calc initial R_n m_n N_n
	float vals[9] = {};
	QMatrix3x3 R(vals);
	QVector3D m(0.0, 0.0, 0.0);

	unsigned int N = imagePixels.size();

	for(int i=0; i < N; i++)
	{
		R = R + multiplyVectors(&imagePixels[i], &imagePixels[i]);
		m += imagePixels[i];
	}

	recursiveSplit(imagePixels, R, m, colorSignatures);

	// sort color signatures
	sort(colorSignatures.begin(), colorSignatures.end(), colorSignatureSortFunction);

	// reorganize color signatures
	signature.n = colorSignatures.size();
	// allocate memory
	float *weights = (float *) malloc(sizeof(float) * signature.n);
	feature_t *features = (feature_t *) malloc(sizeof(feature_t) * signature.n);

	for(int i=0; i < signature.n; i++)
	{
		weights[i] = colorSignatures[i].N / N;
		features[i].r = qRound(colorSignatures[i].pixel.x());
		features[i].g = qRound(colorSignatures[i].pixel.y());
		features[i].b = qRound(colorSignatures[i].pixel.z());
	}

	signature.Weights = weights;
	signature.Features = features;
}

void JunctionDetector::recursiveSplit(
	const vector<QVector3D> &imagePixels,
	const QMatrix3x3 &R, const QVector3D &m,
	vector<ColorSignature> &colorSignatures)
{
	const unsigned int N = imagePixels.size();
	if(N == 0)
		return;

	QVector3D q = m / N;

	if(recursiveDepth > MAX_RECURSIVE_STEPS)
	{
		ColorSignature sig;
		sig.N = N;
		sig.pixel = q;
		colorSignatures.push_back(sig);
		return;
	}

	recursiveDepth++;

	// find out e_n
	QMatrix3x3 R_ = R - (float)(1 / N) * multiplyVectors(&m, &m);
	float lambdas[3];
	QVector3D e_n[3];
	calcCubicEquation(R_, lambdas);

	// break if biggest eigenvalue is too small
	if(lambdas[0] < MIN_EIGENVALUE_VALUE)
	{
		ColorSignature sig;
		sig.N = N;
		sig.pixel = q;
		colorSignatures.push_back(sig);

		recursiveDepth--;
		return;
	}

	eigenvectorSolver(R_, lambdas, e_n);

	vector<vector<QVector3D>> nodes;
	nodes.push_back(vector<QVector3D>());
	nodes.push_back(vector<QVector3D>());

	float thresh = QVector3D::dotProduct(e_n[0], q);
	for(int j=0; j < N; j++)
	{
		if(QVector3D::dotProduct(e_n[0], imagePixels[j])
			<= thresh)
			nodes[0].push_back(imagePixels[j]);
		else
			nodes[1].push_back(imagePixels[j]);
	}

	// calculate values for first node
	unsigned int N1 = nodes[0].size();
	float vals[9] = {};
	QMatrix3x3 R1(vals);
	QVector3D m1(0.0f, 0.0f, 0.0f);

	for(int i=0; i < N1; i++)
	{
		R1 = R1 + multiplyVectors(&nodes[0][i], &nodes[0][i]);
		m1 += nodes[0][i];
	}

	// split recursively
	if(N1 > 0)
		recursiveSplit(nodes[0], R1, m1, colorSignatures);

	// calculate values for second node
	QMatrix3x3 R2 = R - R1;
	QVector3D m2 = m - m1;
	unsigned int N2 = N - N1;
	recursiveSplit(nodes[1], R2, m2, colorSignatures);

	// decrement recursive step
	recursiveDepth--;
}

bool JunctionDetector::colorSignatureSortFunction(
	const ColorSignature &sig1, const ColorSignature &sig2)
{
	return sig1.N > sig2.N;
}

void JunctionDetector::calcCubicEquation(const QMatrix3x3 &A, float root[3])
{
	const float inv3 = 1.0f/3.0f;
	const float root3 = sqrt(3.0f);

	float a00 = (float)A.m[0][0];
	float a01 = (float)A.m[1][0];
	float a02 = (float)A.m[2][0];
	float a11 = (float)A.m[1][1];
	float a12 = (float)A.m[2][1];
	float a22 = (float)A.m[2][2];
	float c0 = a00*a11*a22 + 2.0*a01*a02*a12 - a00*a12*a12 - a11*a02*a02 - a22*a01*a01;
	float c1 = a00*a11 - a01*a01 + a00*a22 - a02*a02 + a11*a22 - a12*a12;
	float c2 = a00 + a11 + a22;
	float c2Div3 = c2*inv3;
	float aDiv3 = (c1 - c2*c2Div3)*inv3;
	if (aDiv3 > 0.0) { aDiv3 = 0.0; }
	float mbDiv2 = 0.5*(c0 + c2Div3*(2.0*c2Div3*c2Div3 - c1));
	float q = mbDiv2*mbDiv2 + aDiv3*aDiv3*aDiv3;
	if (q > 0.0) { q = 0.0; }
	float magnitude = sqrt(-aDiv3);
	float angle = atan2(sqrt(-q),mbDiv2)*inv3;
	float cs = cos(angle);
	float sn = sin(angle);

	float tmp[3];
	tmp[0] = c2Div3 + 2.0*magnitude*cs;
	tmp[1] = c2Div3 - magnitude*(cs + root3*sn);
	tmp[2] = c2Div3 - magnitude*(cs - root3*sn);

	// Sort the roots here to obtain root[0] <= root[1] <= root[2].
	root[0] = qMax(tmp[0], qMax(tmp[1], tmp[2]));
	root[2] = qMin(tmp[0], qMin(tmp[1], tmp[2]));
	for(int i=0; i < 3; i++)
		if(tmp[i] != root[0] && tmp[i] != root[2])
			root[1] = tmp[i];
}

void JunctionDetector::eigenvectorSolver(
	const QMatrix3x3 &A, float evalue[3], QVector3D evector[3])
{
	float root[3];
	calcCubicEquation(A,root);
	evalue[0] = (float)root[0];
	evalue[1] = (float)root[1];
	evalue[2] = (float)root[2];

	QMatrix3x3 M0, M1, M2, I;
	int rank0, rank1, rank2;
	QVector3D row0, row1;
	const QVector3D zero3D(0.0f, 0.0f, 0.0f);

	I.setToIdentity();

	M0 = A - evalue[0]*I; // I is the identity matrix
	//rank0 = computeRank(M0);

	/*
	if (rank0 == 0)
	{
	// evalue[0] = evalue[1] = evalue[2]
	M1 = A - evalue[1]*I;
	M2 = A - evalue[2]*I;
	rank1 = computeRank(M1);
	rank2 = computeRank(M2);
	assert(rank1 == 0 && rank1 == 0);
	evector[0] = QVector3D(1.0f, 0.0f, 0.0f);
	evector[1] = QVector3D(0.0f, 1.0f, 0.0f);
	evector[2] = QVector3D(0.0f, 0.0f, 1.0f);
	return;
	}
	if (rank0 == 1)
	{
	// evalue[0] = evalue[1] < evalue[2]
	M1 = A - evalue[1]*I;
	M2 = A - evalue[2]*I;
	rank1 = computeRank(M1);
	rank2 = computeRank(M2);
	assert(rank1 == 1 && rank1 == 2);
	row0 = QVector3D(M0.m[0][0], M0.m[0][1], M0.m[0][2]); // M row(0)
	row0.normalize();
	assert(row0 != QVector3D(0.0f, 0.0f, 0.0f));
	getComplement2(row0,evector[0],evector[1]);
	evector[2] = QVector3D::crossProduct(evector[0], evector[1]);
	assert(evector[2] != QVector3D(0.0f, 0.0f, 0.0f));
	return;
	}
	*/
	// rank0 == 2
	row0 = QVector3D(M0.m[0][0], M0.m[0][1], M0.m[0][2]); // M row(0)
	row1 = QVector3D(M0.m[1][0], M0.m[1][1], M0.m[1][2]); // M row(0)
	row0.normalize();
	row1.normalize();
	assert(row0 !=  zero3D && row1 != zero3D);
	getComplement1(row0,row1,evector[0]);
	assert(evector[0] != QVector3D(0.0f, 0.0f, 0.0f));
	return;

	/*
	M1 = A - evalue[1]*I;
	rank1 = computeRank(M1);
	assert(rank1 > 0);
	if (rank1 == 1)
	{
	// evalue[0] < evalue[1] = evalue[2]
	M2 = A - evalue[2]*I;
	rank2 = computeRank(M2);
	assert(rank2 == 1);
	getComplement2(evector[0], evector[1], evector[2]);
	return;
	}
	// rank1 == 2
	row0 = QVector3D(M1.m[0][0], M1.m[0][1], M1.m[0][2]); //M1.GetRow(0);
	row1 = QVector3D(M1.m[0][0], M1.m[0][1], M1.m[0][2]); //M1.GetRow(1);
	row0.normalize();
	row1.normalize();
	assert(row0 != QVector3D(0.0f, 0.0f, 0.0f) && row1 != QVector3D(0.0f, 0.0f, 0.0f));
	getComplement1(row0,row1,evector[1]);
	assert(evector[1] != QVector3D(0.0f, 0.0f, 0.0f));
	// rank2 == 2 (eigenvalues must be distinct at this point, rank2 must be 2)
	M2 = A - evalue[2]*I;
	rank2 = computeRank(M2);
	assert(rank2 == 2);
	evector[2] = QVector3D::crossProduct(evector[0], evector[1]);
	assert(evector[2] != QVector3D(0.0f, 0.0f, 0.0f));
	*/
}

int JunctionDetector::computeRank(QMatrix3x3 M)
{
	// Compute the maximum magnitude matrix entry.
	float absVal, save, max = -1;
	int row, col, maxRow = -1, maxCol = -1;
	for (row = 0; row < 3; row++)
	{
		for (col = row; col < 3; col++)
		{
			absVal = abs(M.m[col][row]);
			if (absVal > max)
			{
				max = absVal;
				maxRow = row;
				maxCol = col;
			}
		}
	}
	if (max == 0)
	{
		// The rank is 0. The eigenvalue has multiplicity 3.
		return 0;
	}
	// The rank is at least 1. Swap the row containing the maximum-magnitude
	// entry with row 0.
	if (maxRow != 0)
	{
		for (col = 0; col < 3; col++)
		{
			save = M.m[col][0];
			M.m[col][0] = M.m[col][maxRow];
			M.m[col][maxRow] = save;
		}
	}
	// Row-reduce the matrix...
	// Scale row 0 to generate a 1-valued pivot.
	float invMax = 1/M.m[maxCol][0];
	M.m[0][0] *= invMax;
	M.m[1][0] *= invMax;
	M.m[2][0] *= invMax;
	// Eliminate the maxCol column entries in rows 1 and 2.
	if (maxCol == 0)
	{
		M.m[1][1] -= M.m[0][1]*M.m[1][0];
		M.m[2][1] -= M.m[0][1]*M.m[2][0];
		M.m[1][2] -= M.m[0][2]*M.m[1][0];
		M.m[2][2] -= M.m[0][2]*M.m[2][0];
		M.m[0][1] = 0;
		M.m[0][2] = 0;
	}
	else if (maxCol == 1)
	{
		M.m[0][1] -= M.m[1][1]*M.m[0][0];
		M.m[2][1] -= M.m[1][1]*M.m[2][0];
		M.m[0][2] -= M.m[1][2]*M.m[0][0];
		M.m[2][2] -= M.m[1][2]*M.m[2][0];
		M.m[1][1] = 0;
		M.m[2][1] = 0;
	}
	else
	{
		M.m[0][1] -= M.m[1][2]*M.m[0][0];
		M.m[1][1] -= M.m[1][2]*M.m[1][0];
		M.m[0][2] -= M.m[2][2]*M.m[0][0];
		M.m[1][2] -= M.m[2][2]*M.m[1][0];
		M.m[2][1] = 0;
		M.m[2][2] = 0;
	}
	// Compute the maximum-magnitude entry of the last two rows of the
	// row-reduced matrix.
	max = -1;
	maxRow = -1;
	maxCol = -1;
	for (row = 1; row < 3; row++)
	{
		for (col = 0; col < 3; col++)
		{
			absVal = abs(M.m[row][col]);
			if (absVal > max)
			{
				max = absVal;
				maxRow = row;
				maxCol = col;
			}
		}
	}
	if (max == 0)
	{
		// The rank is 1. The eigenvalue has multiplicity 2.
		return 1;
	}
	// If row 2 has the maximum-magnitude entry, swap it with row 1.
	if (maxRow == 2)
	{
		for (col = 0; col < 3; col++)
		{
			save = M.m[col][1];
			M.m[col][2] = M.m[col][2];
			M.m[col][2] = save;
		}
	}
	// Scale row 1 to generate a 1-valued pivot.
	invMax = 1/M.m[maxCol][1];
	M.m[0][1] *= invMax;
	M.m[1][1] *= invMax;
	M.m[2][1] *= invMax;

	// The rank is 2. The eigenvalue has multiplicity 1.
	return 2;
}

void JunctionDetector::getComplement1 (const QVector3D &U, const QVector3D &V, QVector3D &W)
{
	W = QVector3D::crossProduct(U, V);
	W.normalize();
}

void JunctionDetector::getComplement2 (
	QVector3D U, QVector3D &V, QVector3D &W)
{
	U.normalize();
	if (abs(U.x()) >= abs(U.y()))
	{
		float invLength = 1/sqrt(U.x()*U.x() + U.z()*U.z());
		V.setX(-U.z()*invLength);
		V.setY(0.0f);
		V.setZ(U.x()*invLength);
		W.setX(U.y()*V.z());
		W.setY(U.z()*V.x() - U.x()*V.z());
		W.setZ(-U.y()*V.x());
	}
	else
	{
		float invLength = 1/sqrt(U.y()*U.y() + U.z()*U.z());
		V.setX(0);
		V.setY(U.z()*invLength);
		V.setZ(-U.y()*invLength);
		W.setX(U.y()*V.z() - U.z()*V.y());
		W.setY(-U.x()*V.z());
		W.setZ(U.x()*V.y());
	}
}