#ifndef JUNCTIONDETECTOR_H
#define JUNCTIONDETECTOR_H

#include <QtGui>
#include <vector>
#include <cmath>

extern "C" {
	#include "emd.h"
}

#define MAX_RECURSIVE_STEPS 1
#define MIN_EIGENVALUE_VALUE 0.1f

using namespace std;

/**
 * JunctionDetecotr class.
 * This class provides all the necessary methods to do X junction detection.
 * @author Christian Hirsch
 * @author Clemens Roegner
 */
class JunctionDetector
{
public:

	/**
	 * Detect junctions in an image.
	 * This method computes the junctions in an image using the images gradient fields and canny edges.
	 * @param image The original image used to compute junctions.
	 * @param cannyEdgeImage The image in which the canny edges are drawn (R, G, B = 255, 0, 0).
	 * @param gradientImage The image in which the gradients are drawn (R, G = gradient vector, B = magnitude).
	 * @param junctionImage In this image the calculated junctions are drawn (R, G, B = 0, 255, 0).
	 * @param outputJunctionImage In this image, the junctions are marked.
	 */
	static void detectJunctions(const QImage &image,
		const QImage &cannyEdgeImage,
		const QImage &gradientImage,
		QImage &junctionImage,
		QImage &outputJunctionImage);

	/**
	 * Detect X-junctions of an image.
	 * This method calculates the X-junctions from an image and stores them in another image. It also returns the surrounding luminances of an X-junction.
	 * @param image The original image used to calculate the X-junctions.
	 * @param junctionImage This image originally consists of all the junctions (R, G, B = 0, 255, 0) and the calculated X-junctions are drawn into it.
	 * @param outputXJunctionImage In this image, the X-junctions are marked.
	 * @param luminances Stores the surrounding 4 luminances of an detected X-junction.
	 */
	static void detectXJunctions(
		QImage &image,
		QImage &junctionImage,
		QImage &outputXJunctionImage,
		std::vector<std::vector<float>> &luminances);

	/**
	 * Color signature.
	 * Stores a pixel value and its weight.
	 */
	struct ColorSignature {
		float N;
		QVector3D pixel;
	};


private:
	static void calculateColorsignature(
		const vector<QVector3D> &imagePixels,
		signature_t &signature);

	static void recursiveSplit(
		const vector<QVector3D> &imagePixels,
		const QMatrix3x3 &R, const QVector3D &m,
		vector<ColorSignature> &colorSignatures);

	static QMatrix3x3 multiplyVectors(
		const QVector3D *vec,
		const QVector3D *vecT);
	
	static void calcCubicEquation(
		const QMatrix3x3 &A,
		float root[3]);
	
	static void eigenvectorSolver(
		const QMatrix3x3 &A,
		float evalue[3],
		QVector3D res[3]);
	
	static int computeRank(QMatrix3x3 M);
	
	static void getComplement1 (
		const QVector3D &U,
		const QVector3D &V,
		QVector3D &W);
	
	static void getComplement2 (
		QVector3D U,
		QVector3D &V,
		QVector3D &W);

	static bool colorSignatureSortFunction(
		const ColorSignature &sig1,
		const ColorSignature &sig2);

	static void growRegion(
		const int &seedX, const int &seedY,
		QImage &image,
		const QColor &initialColor,
		QImage &neighbours);

	static float colorDiff(
		const QColor &color1,
		const QColor &color2);

	static unsigned int recursiveDepth;
	static const QVector2D normalizedVector;
};

#endif