#include "renderwindow.h"

RenderWindow::RenderWindow(QScreen *screen)
	:QWindow(screen)
{
	setMinimumSize(QSize(700, 700));

	// Tell Qt we will use OpenGL for this window
    setSurfaceType( OpenGLSurface );
 
    // Specify the format and create platform-specific surface
	QSurfaceFormat format(QSurfaceFormat::DebugContext);
    format.setDepthBufferSize( 24 );
    format.setMajorVersion( 3 );
    format.setMinorVersion( 2 );
    format.setSamples( 4 );
    format.setProfile( QSurfaceFormat::CoreProfile );
    setFormat( format );
    create();
 
    // Create an OpenGL context
	v_context = new QOpenGLContext();
	v_context->setFormat(format);
	v_context->create();

	// Make the context current on this window
	v_context->makeCurrent(this);

	// Obtain a functions object and resolve all entry points
    // m_funcs is declared as: QOpenGLFunctions_3_2_Core* v_funcs
    v_funcs = v_context->versionFunctions<QOpenGLFunctions_3_2_Core>();
    if ( !v_funcs ) {
        qCritical() << "Could not obtain OpenGL versions object";
        exit( 1 );
    }
    v_funcs->initializeOpenGLFunctions();

	volumeData = new Volume;

	cannyEdgeDetector = new CannyEdge(
		this, v_context,
		1.4f, 7,
		width(), height());

	initializeSettings();
	initializeMatrices();
	initialize();
	resize(width(), height());
}

RenderWindow::~RenderWindow()
{
	delete volumeData;
}

void RenderWindow::initialize()
{
	v_context->makeCurrent(this);
	
	m_logger = new QOpenGLDebugLogger( this );
 
    connect( m_logger, SIGNAL( messageLogged( QOpenGLDebugMessage ) ),
             this, SLOT( onMessageLogged( QOpenGLDebugMessage ) ),
             Qt::DirectConnection );
 
    if ( m_logger->initialize() ) {
        m_logger->startLogging( QOpenGLDebugLogger::SynchronousLogging );
        m_logger->enableMessages();
    }

	// Enable depth testing
    v_funcs->glEnable( GL_DEPTH_TEST );
	v_funcs->glEnable( GL_CULL_FACE );
	v_funcs->glEnable(GL_BLEND);
	v_funcs->glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	v_funcs->glViewport(0, 0, width(), height());
	v_funcs->glClearColor(1.0f, 1.0f, 1.0f, 1.0f);

	prepareTextures();
	prepareShaders();
	prepareFramebuffers();
	prepareQuadVertexArrayObject();
	prepareCubeVertexArrayObject();

	show();
	
	vMatrixVolume.setToIdentity();
	vMatrixVolume.lookAt(QVector3D(0.0f, 0.0f, distanceView),
		QVector3D(0.0f, 0.0f, 0.0f),
		QVector3D(0.0f, 1.0f, 0.0f));
	
	mvpMatrixScreenRect.setToIdentity();
	mvpMatrixScreenRect.ortho(-1.0, 1.0, -1.0, 1.0, -1.0, 1.0);

	v_funcs->glGenTextures(1, &v_tmpTex);

	updateAndRender();
}

void RenderWindow::updateAndRender()
{
	update();
	render();
}

void RenderWindow::update()
{
	if(renderType == RENDER_SLICE)
		updateSlice();
	else if(renderType == RENDER_VOLUME)
		updateVolume();
}

void RenderWindow::updateSlice()
{
	v_sliceShader->bind();
	v_sliceShader->setUniformValue("mvpMatrix", mvpMatrixScreenRect);
}

void RenderWindow::updateVolume()
{
	junctionImage.fill(QColor(0, 0, 0, 0));
	xJunctionImage.fill(QColor(0, 0, 0, 0));
	QMatrix4x4 cameraTransformation;
	cameraTransformation.rotate(alphaView, 0.0f, 1.0f, 0.0f);
	cameraTransformation.rotate(betaView, 1.0f, 0.0f, 0.0f);
	
	QVector3D cameraPosition = cameraTransformation * QVector3D(0.0f, 0.0f, distanceView);
	QVector3D cameraUpDirection = cameraTransformation * QVector3D(0.0f, 1.0f, 0.0f);
	
	vMatrixVolume.setToIdentity();
	vMatrixVolume.lookAt(cameraPosition, QVector3D(0.0f, 0.0f, 0.0f), cameraUpDirection);
	pMatrixVolume.setToIdentity();

	if(projectionType == PERSPECTIVE)
	{
		pMatrixVolume.perspective(60.0f, (float) width()/height(), 0.000001f, 2048.0f);
		
	}

	else if(projectionType == ORTHOGONAL)
	{
		pMatrixVolume.ortho(-1.0f, 1.0f, -1.0f, 1.0f, 0.000001f, 2048.0f);
		vMatrixVolume.scale(distanceView);
	}

	renderCubeBackfaces();
	renderCubeFrontfaces();
}

void RenderWindow::render()
{
	if(renderType == RENDER_SLICE)
		renderSlice();
	else if(renderType == RENDER_VOLUME)
		renderVolume();
}

void RenderWindow::renderSlice()
{
	v_context->makeCurrent(this);

	// Clear buffers
	v_funcs->glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Bind textures
	v_funcs->glActiveTexture(GL_TEXTURE2);
	v_funcs->glBindTexture(GL_TEXTURE_3D, v_sliceTex);
	v_funcs->glActiveTexture(GL_TEXTURE3);
	v_funcs->glBindTexture(GL_TEXTURE_1D, v_transferfunctionTex);

	// Bind shader
	v_sliceShader->bind();
	v_sliceShader->setUniformValue("mvpMatrix", mvpMatrixScreenRect);
	v_sliceShader->setUniformValue("sliceTexture", GLint(2));
	v_sliceShader->setUniformValue("transferFunction", GLint(3));
	v_sliceShader->setUniformValue("sliceNumber", sliceNumber);
	v_sliceShader->setUniformValue("sliceMatrix", vMatrixSlice);

	// Switch to vertex data
	QOpenGLVertexArrayObject::Binder binder( &v_quad );
	v_funcs->glDrawArrays(GL_TRIANGLES, 0, 6);
	
	v_sliceShader->release();
	v_context->swapBuffers(this);
}

void RenderWindow::renderVolume()
{
	v_screenRectFB->bind();

	v_funcs->glDisable(GL_CULL_FACE);
	v_funcs->glEnable(GL_BLEND);
	v_funcs->glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_DST_ALPHA);

	// Bind textures
	v_funcs->glActiveTexture(GL_TEXTURE0);
	v_funcs->glBindTexture(GL_TEXTURE_2D, v_frontFB->texture());
	v_funcs->glActiveTexture(GL_TEXTURE1);
	v_funcs->glBindTexture(GL_TEXTURE_2D, v_backFB->texture());

	// Bind shader
	v_rayShader->bind();
	v_rayShader->setUniformValue("mvpMatrix", mvpMatrixScreenRect);
	v_rayShader->setUniformValue("frontTexture", GLint(0));
	v_rayShader->setUniformValue("backTexture", GLint(1));
	v_rayShader->setUniformValue("volumeData", GLint(2));
	v_rayShader->setUniformValue("transferFunction", GLint(3));
	v_rayShader->setUniformValue("mode", volumeRenderMode);
	v_rayShader->setUniformValue("dz", volumeStepsize);
	v_rayShader->setUniformValue("light", volumeShadedRendering);
	v_rayShader->setUniformValue("compMode", volumeCompositionMode);
	v_rayShader->setUniformValue("vMatrix", vMatrixVolume * mMatrixVolume);

	// Clear buffers
	v_funcs->glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Switch to vertex data
	QOpenGLVertexArrayObject::Binder binderQuad( &v_cube );
	v_funcs->glDrawArrays(GL_TRIANGLES, 0, 6);

	v_screenRectFB->release();
	v_rayShader->release();

	cannyEdgeDetector->detectEdges(v_screenRectFB->texture());
	renderTextureRectToScreen(v_screenRectFB->texture());
	if(showCannyEdges)
		renderTextureRectToScreen(cannyEdgeDetector->getCannyEdgeTexture(), false);
	if(showJunctions)
		renderQImageToScreen(junctionImage, false);
	if(showXJunctions)
		renderQImageToScreen(xJunctionImage, false);
}

void RenderWindow::calculateDepthEnergy()
{
	QImage tmpJunctionImage(width(), height(), QImage::Format_ARGB32);

	// detect junctions
	JunctionDetector::detectJunctions(
		v_screenRectFB->toImage(),
		cannyEdgeDetector->getCannyEdgeImage(),
		cannyEdgeDetector->getGradientFieldImage(),
		tmpJunctionImage,
		junctionImage);

	vector<vector<float>> luminances;

	// detect X-junctions
	QImage screenCapt = flipImageHorizontally(v_screenRectFB->toImage());
	JunctionDetector::detectXJunctions(
		screenCapt,
		//cannyEdgeDetector->getGradientFieldImage(),
		tmpJunctionImage,
		xJunctionImage,
		luminances);

	render();

	// calculate depth ordering
	depthOrdering = 0.0f;
	for(int i=0; i < luminances.size(); i++)
	{
		float ed1 = 0.75f + 0.25f * tanh(luminances[i][3] - luminances[i][2]);
		float d2 = qSqrt((
			qPow(qAbs(luminances[i][0] - luminances[i][1]), 2)
			+ qPow(qAbs(luminances[i][1] - luminances[i][2]), 2)
			+ qPow(qAbs(luminances[i][2] - luminances[i][3]), 2)
			) / 3.0f);
		float ed2 = 0.5f * tanh(1.0f/d2 - 1);

		if(luminances[i][3] >= luminances[i][1])
			depthOrdering += 1.0f / luminances.size() * ed1;
		else
			depthOrdering += 1.0f / luminances.size() * ed2;
	}
	emit depthOrderChanged(depthOrdering);
}

void RenderWindow::renderQImageToScreen(
	const QImage &image, bool clearScreen)
{
	// render junctions
	v_funcs->glBindTexture(GL_TEXTURE_RECTANGLE, v_tmpTex);
	v_funcs->glTexImage2D(GL_TEXTURE_RECTANGLE, 0, GL_RGBA8,
		image.width(), image.height(),
		0, GL_BGRA, GL_UNSIGNED_BYTE,
		image.bits());

	renderTextureRectToScreen(v_tmpTex, clearScreen);
}

void RenderWindow::renderCubeBackfaces()
{
	v_funcs->glEnable(GL_DEPTH_TEST);
	v_funcs->glEnable(GL_CULL_FACE);

	// Bind shader
	v_cubeShader->bind();
	v_cubeShader->setUniformValue("mvpMatrix", pMatrixVolume * vMatrixVolume * mMatrixVolume);

	// Bind framebuffers
	v_backFB->bind();
	if(!v_backFB->isBound())
		qDebug() << "QOpenGLFrameBufferObject \'v_backFB\' is not bound.";

	// Clear buffers
	v_funcs->glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	// Switch to vertex data
	v_funcs->glCullFace(GL_FRONT);
	QOpenGLVertexArrayObject::Binder binderCube( &v_cube );
	v_funcs->glDrawArrays(GL_TRIANGLES, 0, 36);
	v_backFB->release();

	v_cubeShader->release();
}

void RenderWindow::renderCubeFrontfaces()
{
	v_funcs->glEnable(GL_DEPTH_TEST);
	v_funcs->glEnable(GL_CULL_FACE);

	// Bind shader
	v_cubeShader->bind();
	v_cubeShader->setUniformValue("mvpMatrix", pMatrixVolume * vMatrixVolume * mMatrixVolume);

	// Bind framebuffers
	v_frontFB->bind();
	if(!v_frontFB->isBound())
		qDebug() << "QOpenGLFrameBufferObject \'v_frontFB\' is not bound.";

	// Clear buffers
	v_funcs->glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	// Switch to vertex data
	v_funcs->glCullFace(GL_BACK);
	QOpenGLVertexArrayObject::Binder binderCube( &v_cube );
	v_funcs->glDrawArrays(GL_TRIANGLES, 0, 36);
	v_frontFB->release();

	v_cubeShader->release();
}

void RenderWindow::renderTextureToScreen(GLuint texture, bool clearScreen)
{
	v_context->makeCurrent(this);

	v_funcs->glDisable(GL_CULL_FACE);
	v_funcs->glDisable(GL_DEPTH_TEST);
	v_funcs->glEnable(GL_BLEND);
	v_funcs->glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	// Clear buffers
	if(clearScreen)
		v_funcs->glClear(GL_COLOR_BUFFER_BIT);

	// Bind textures
	v_funcs->glActiveTexture(GL_TEXTURE0);
	v_funcs->glBindTexture(GL_TEXTURE_2D, texture);

	// Bind shader
	v_imageShader->bind();
	v_imageShader->setUniformValue("image", GLint(0));

	// Switch to vertex data
	QOpenGLVertexArrayObject::Binder binder( &v_quad );
	v_funcs->glDrawArrays(GL_TRIANGLES, 0, 6);
	
	v_imageShader->release();
	v_context->swapBuffers(this);
}

void RenderWindow::renderTextureRectToScreen(GLuint texture, bool clearScreen)
{
	v_context->makeCurrent(this);

	v_funcs->glDisable(GL_CULL_FACE);
	v_funcs->glDisable(GL_DEPTH_TEST);
	v_funcs->glEnable(GL_BLEND);
	v_funcs->glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	// Clear buffers
	if(clearScreen)
		v_funcs->glClear(GL_COLOR_BUFFER_BIT);

	// Bind textures
	v_funcs->glActiveTexture(GL_TEXTURE0);
	v_funcs->glBindTexture(GL_TEXTURE_RECTANGLE, texture);

	// Bind shader
	v_imageRectShader->bind();
	v_imageRectShader->setUniformValue("width", (float)width());
	v_imageRectShader->setUniformValue("height", (float)height());
	v_imageRectShader->setUniformValue("image", GLint(0));

	// Switch to vertex data
	QOpenGLVertexArrayObject::Binder binder( &v_quad );
	v_funcs->glDrawArrays(GL_TRIANGLES, 0, 6);
	
	v_imageRectShader->release();
	v_context->swapBuffers(this);
}

void RenderWindow::resize(int width, int height) {
	v_funcs->glViewport(0, 0, width, height);
	
	// recreate framebuffers
	delete(v_backFB);
	delete(v_frontFB);
	prepareFramebuffers();

	delete(cannyEdgeDetector);
	cannyEdgeDetector = new CannyEdge(
		this, v_context,
		2.5f, 15,
		width, height);

	junctionImage = QImage(width, height, QImage::Format_ARGB32);
	junctionImage.fill(QColor(0, 0, 0, 0));
	xJunctionImage = QImage(width, height, QImage::Format_ARGB32);
	xJunctionImage.fill(QColor(0, 0, 0, 0));

	updateAndRender();
}

void RenderWindow::prepareShaders()
{
	v_sliceShader = new QOpenGLShaderProgram();
	if(!v_sliceShader->addShaderFromSourceFile(QOpenGLShader::Vertex, "res/SliceShader.vert"))
		qCritical() << "Could not compile vertex shader. Log: " << v_sliceShader->log();
	if(!v_sliceShader->addShaderFromSourceFile(QOpenGLShader::Fragment, "res/SliceShader.frag"))
		qCritical() << "Could not compile fragment shader. Log: " << v_sliceShader->log();
	v_sliceShader->link();

	v_cubeShader = new QOpenGLShaderProgram();
	if(!v_cubeShader->addShaderFromSourceFile(QOpenGLShader::Vertex, "res/CubeShader.vert"))
		qCritical() << "Could not compile vertex shader. Log: " << v_cubeShader->log();
	if(!v_cubeShader->addShaderFromSourceFile(QOpenGLShader::Fragment, "res/CubeShader.frag"))
		qCritical() << "Could not compile fragment shader. Log: " << v_cubeShader->log();
	v_cubeShader->link();

	v_rayShader = new QOpenGLShaderProgram();
	if(!v_rayShader->addShaderFromSourceFile(QOpenGLShader::Vertex, "res/RayShader.vert"))
		qCritical() << "Could not compile vertex shader. Log: " << v_rayShader->log();
	if(!v_rayShader->addShaderFromSourceFile(QOpenGLShader::Fragment, "res/RayShader.frag"))
		qCritical() << "Could not compile fragment shader. Log: " << v_rayShader->log();
	v_rayShader->link();

	v_imageRectShader = new QOpenGLShaderProgram();
	if(!v_imageRectShader->addShaderFromSourceFile(QOpenGLShader::Vertex, "res/ImageRectShader.vert"))
		qCritical() << "Could not compile vertex shader. Log: " << v_imageRectShader->log();
	if(!v_imageRectShader->addShaderFromSourceFile(QOpenGLShader::Fragment, "res/ImageRectShader.frag"))
		qCritical() << "Could not compile fragment shader. Log: " << v_imageRectShader->log();
	v_imageRectShader->link();

	v_imageShader = new QOpenGLShaderProgram();
	if(!v_imageShader->addShaderFromSourceFile(QOpenGLShader::Vertex, "res/ImageShader.vert"))
		qCritical() << "Could not compile vertex shader. Log: " << v_imageShader->log();
	if(!v_imageShader->addShaderFromSourceFile(QOpenGLShader::Fragment, "res/ImageShader.frag"))
		qCritical() << "Could not compile fragment shader. Log: " << v_imageShader->log();
	v_imageShader->link();
}

void RenderWindow::prepareTextures()
{
	v_funcs->glGenTextures(1, &v_sliceTex);
	v_funcs->glGenTextures(1, &v_transferfunctionTex);
}

void RenderWindow::prepareFramebuffers()
{
	v_context->makeCurrent(this);

	QOpenGLFramebufferObjectFormat format;
	format.setAttachment(QOpenGLFramebufferObject::CombinedDepthStencil);
	format.setMipmap(false);
	format.setTextureTarget(GL_TEXTURE_2D);
	format.setInternalTextureFormat(GL_RGBA16);
	

	v_frontFB = new QOpenGLFramebufferObject(
		width(), height(), format);
	if(!v_frontFB->isValid())
		qDebug() << "Failed to create QOpenGLFramebufferObject \'v_frontFB\'";

	v_funcs->glBindTexture(GL_TEXTURE_2D, v_frontFB->texture());
	v_funcs->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	v_funcs->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	v_funcs->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	v_funcs->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);


	v_backFB = new QOpenGLFramebufferObject(
		width(), height(), format);
	if(!v_backFB->isValid())
		qDebug() << "Failed to create QOpenGLFramebufferObject \'v_backFB\'";

	v_funcs->glBindTexture(GL_TEXTURE_2D, v_backFB->texture());
	v_funcs->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	v_funcs->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	v_funcs->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	v_funcs->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);


	format.setTextureTarget(GL_TEXTURE_RECTANGLE);
	v_screenRectFB = new QOpenGLFramebufferObject(
		width(), height(), format);

	v_funcs->glBindTexture(GL_TEXTURE_RECTANGLE, v_backFB->texture());
	v_funcs->glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	v_funcs->glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	v_funcs->glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	v_funcs->glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
}

void RenderWindow::prepareQuadVertexArrayObject()
{
	const int vertexCount = 6;

	// Setup VBOs and IBO (use QOpenGLBuffer to buffer data,
    // specify format, usage hint etc). These will be
    // remembered by the currently bound VAO
	float vertexPositions[] = {
		-1.0f,  1.0f,  0.0f,
		-1.0f, -1.0f,  0.0f,
		 1.0f, -1.0f,  0.0f,
		 1.0f, -1.0f,  0.0f,
		 1.0f,  1.0f,  0.0f,
		-1.0f,  1.0f,  0.0f
	};
	QOpenGLBuffer v_positionBuffer;
    v_positionBuffer.create();
	v_positionBuffer.setUsagePattern( QOpenGLBuffer::StaticDraw );
    v_positionBuffer.bind();
	v_positionBuffer.allocate(
		vertexPositions,
        vertexCount * 3 * sizeof(float)
		);
	v_positionBuffer.release();

	float texCoords[] = {
		0.0f, 1.0f,
		0.0f, 0.0f,
		1.0f, 0.0f,
		1.0f, 0.0f,
		1.0f, 1.0f,
		0.0f, 1.0f
	};
	QOpenGLBuffer v_texCoordsBuffer;
	v_texCoordsBuffer.create();
	v_positionBuffer.setUsagePattern( QOpenGLBuffer::StaticDraw );
	v_texCoordsBuffer.bind();
	v_texCoordsBuffer.allocate(
		texCoords,
		vertexCount * 2 * sizeof(float)
		);
	v_texCoordsBuffer.release();

    // Create VAO for first object to render
    v_quad.create();
	QOpenGLVertexArrayObject::Binder binder( &v_quad );

	v_positionBuffer.bind();
	v_rayShader->enableAttributeArray("inputVertexPosition");
	v_rayShader->setAttributeBuffer("inputVertexPosition", GL_FLOAT, 0, 3);
	v_positionBuffer.release();
	v_rayShader->enableAttributeArray("inputTexCoord");
	v_rayShader->setAttributeBuffer("inputTexCoord", GL_FLOAT, 0, 2);
	v_texCoordsBuffer.release();
	
	binder.rebind();
	v_positionBuffer.bind();
	v_sliceShader->enableAttributeArray("inputVertexPosition");
	v_sliceShader->setAttributeBuffer("inputVertexPosition", GL_FLOAT, 0, 3);
	v_positionBuffer.release();
	v_texCoordsBuffer.bind();
	v_sliceShader->enableAttributeArray("inputTexCoord");
	v_sliceShader->setAttributeBuffer("inputTexCoord", GL_FLOAT, 0, 2);
	v_texCoordsBuffer.release();

	binder.rebind();
	v_positionBuffer.bind();
	v_imageRectShader->enableAttributeArray("inputVertexPosition");
	v_imageRectShader->setAttributeBuffer("inputVertexPosition", GL_FLOAT, 0, 3);
	v_positionBuffer.release();

	binder.release();
}

void RenderWindow::prepareCubeVertexArrayObject()
{
	const int vertexCount = 6 * 6;

	// Setup VBOs and IBO (use QOpenGLBuffer to buffer data,
    // specify format, usage hint etc). These will be
    // remembered by the currently bound VAO
	float vertexPositions[] = {
		// front face
		-1.0f,  1.0f,  1.0f,
		-1.0f, -1.0f,  1.0f,
		 1.0f, -1.0f,  1.0f,
		 1.0f, -1.0f,  1.0f,
		 1.0f,  1.0f,  1.0f,
		-1.0f,  1.0f,  1.0f,

		// back face
		 1.0f, -1.0f, -1.0f,
		-1.0f, -1.0f, -1.0f,
		-1.0f,  1.0f, -1.0f,
		-1.0f,  1.0f, -1.0f,
		 1.0f,  1.0f, -1.0f,
		 1.0f, -1.0f, -1.0f,

		 // top face
		 1.0f,  1.0f, -1.0f,
		-1.0f,  1.0f, -1.0f,
		-1.0f,  1.0f,  1.0f,
		-1.0f,  1.0f,  1.0f,
		 1.0f,  1.0f,  1.0f,
		 1.0f,  1.0f, -1.0f,

		// bottom face
		-1.0f, -1.0f,  1.0f,
		-1.0f, -1.0f, -1.0f,
		 1.0f, -1.0f, -1.0f,
		 1.0f, -1.0f, -1.0f,
		 1.0f, -1.0f,  1.0f,
		-1.0f, -1.0f,  1.0f,

		// left face
		-1.0f,  1.0f, -1.0f,
		-1.0f, -1.0f, -1.0f,
		-1.0f, -1.0f,  1.0f,
		-1.0f, -1.0f,  1.0f,
		-1.0f,  1.0f,  1.0f,
		-1.0f,  1.0f, -1.0f,

		// right face
		 1.0f, -1.0f,  1.0f,
		 1.0f, -1.0f, -1.0f,
		 1.0f,  1.0f, -1.0f,
		 1.0f,  1.0f, -1.0f,
		 1.0f,  1.0f,  1.0f,
		 1.0f, -1.0f,  1.0f,
	};

	QOpenGLBuffer v_positionBuffer;
    v_positionBuffer.create();
	v_positionBuffer.setUsagePattern( QOpenGLBuffer::StaticDraw );
    v_positionBuffer.bind();
	v_positionBuffer.allocate(
		vertexPositions,
        36 * 3 * sizeof(float)
		);
	v_positionBuffer.release();

	// Create VAO for first object to render
	v_cube.create();
	QOpenGLVertexArrayObject::Binder binder( &v_cube );

	v_positionBuffer.bind();
	v_cubeShader->enableAttributeArray("inputVertexPosition");
	v_cubeShader->setAttributeBuffer("inputVertexPosition", GL_FLOAT, 0, 3);
	v_positionBuffer.release();
}

void RenderWindow::onMessageLogged( QOpenGLDebugMessage message )
{
    qDebug() << message;
}

void RenderWindow::onTimeout()
{
	render();
}

void RenderWindow::setVolumeData(const std::string & strFilename)
{
	volumeData->load(strFilename);

	v_funcs->glActiveTexture(GL_TEXTURE2);

	v_funcs->glBindTexture(GL_TEXTURE_3D, v_sliceTex);

	v_funcs->glTexImage3D(
		GL_TEXTURE_3D, 0,
		GL_RED,
		volumeData->GetWidth(),
		volumeData->GetHeight(),
		volumeData->GetDepth(),
		0, GL_RED, GL_FLOAT, (void *) volumeData->Get());

	// linear interpolation
	v_funcs->glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	v_funcs->glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	// clamp to edge
	v_funcs->glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
	v_funcs->glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
	v_funcs->glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);

	// disable mipmapping
	v_funcs->glTexParameteri(GL_TEXTURE_3D, GL_GENERATE_MIPMAP, false);

	mMatrixVolume.setToIdentity();
	int maxEdge = qMax(qMax(volumeData->GetWidth(), volumeData->GetHeight()), volumeData->GetDepth());
	mMatrixVolume.scale(
		(float)volumeData->GetWidth() / (float)maxEdge,
		(float)volumeData->GetHeight() / (float)maxEdge,
		(float)volumeData->GetDepth() / (float)maxEdge);

	updateAndRender();
}

// TODO: do memory handling!
void RenderWindow::setTransferfunction(const vector<TransferWidget::TransferFeature> &features)
{

	float transferfunctionValues[4096 * 4] = {};

	if(useBluredTF)
	{
		for(int i=0; i < 4096; i++)
		{
			for(int j=0; j < features.size(); j++)
			{
				float x = (float) i / 4096;
				float sig = features[j].sigma;
				float maxAlpha = 1 / (qSqrt(2 * PI) * sig);
				float alpha = qExp(-0.5f*qPow((x - features[j].pos) * 10.0f / sig, 2)) / (qSqrt(2 * PI) * sig);
				alpha *= features[j].color.alphaF() / maxAlpha;

				transferfunctionValues[i * 4 + 3] += alpha;
				if(!(transferfunctionValues[i * 4 + 3] > 0))
					continue;

				float dAlpha = alpha / transferfunctionValues[i * 4 + 3];
				transferfunctionValues[i * 4 + 0] = transferfunctionValues[i * 4 + 0] * (1 - dAlpha)
					+ dAlpha * features[j].color.redF();
				transferfunctionValues[i * 4 + 1] = transferfunctionValues[i * 4 + 1] * (1 - dAlpha)
					+ dAlpha * features[j].color.greenF();
				transferfunctionValues[i * 4 + 2] = transferfunctionValues[i * 4 + 2] * (1 - dAlpha)
					+ dAlpha * features[j].color.blueF();
			}
		}
	} else
	{
		for(int i=1; i < features.size(); i++)
		{
			int start = features[i-1].pos * 4096;
			int end = features[i].pos * 4096;

			float deltaR = features[i].color.redF() - features[i-1].color.redF();
            deltaR /= (end-start);
            float deltaG = features[i].color.greenF() - features[i-1].color.greenF();
            deltaG /= (end-start);
            float deltaB = features[i].color.blueF() - features[i-1].color.blueF();
            deltaB /= (end-start);
            float deltaA = features[i].color.alphaF() - features[i-1].color.alphaF();
            deltaA /= (end-start);

            // red
            transferfunctionValues[start*4 + 0] = features[i-1].color.redF();
            // green
            transferfunctionValues[start*4 + 1] = features[i-1].color.greenF();
            // blue
            transferfunctionValues[start*4 + 2] = features[i-1].color.blueF();
            //alpha
            transferfunctionValues[start*4 + 3] = features[i-1].color.alphaF();

            for(int j = start+1; j < end; j++)
            {
                    // red
                    transferfunctionValues[j*4 + 0] = transferfunctionValues[(j-1)*4 + 0] + deltaR;
                    // green
                    transferfunctionValues[j*4 + 1] = transferfunctionValues[(j-1)*4 + 1] + deltaG;
                    // blue
                    transferfunctionValues[j*4 + 2] = transferfunctionValues[(j-1)*4 + 2] + deltaB;
                    //alpha
                    transferfunctionValues[j*4 + 3] = transferfunctionValues[(j-1)*4 + 3] + deltaA;

                    //printf("%f/%f/%f/%f\n", transferfunctionValues[j*4 + 0], transferfunctionValues[j*4 + 1], transferfunctionValues[j*4 + 2], transferfunctionValues[j*4 + 3]);
            }
		}
	}

	v_funcs->glBindTexture(GL_TEXTURE_1D, v_transferfunctionTex);

	v_funcs->glTexImage1D(GL_TEXTURE_1D, 0,
		GL_RGBA, 4096, 0,
		GL_RGBA, GL_FLOAT,
		transferfunctionValues);

	v_funcs->glTexParameteri(GL_TEXTURE_1D,
		GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	v_funcs->glTexParameteri(GL_TEXTURE_1D,
		GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	
	updateAndRender();
}

void RenderWindow::updateSlice(int _currentSlice)
{
	sliceNumber = (float)_currentSlice / 1023.0f;

	updateAndRender();
}

void RenderWindow::setSliceDirection(int _sliceDirection)
{
	sliceDirection = _sliceDirection;
	vMatrixSlice.setToIdentity();
	switch(sliceDirection)
	{
	case 0: // X
		vMatrixSlice.rotate(180.0f, 1.0f, 0.0f, 0.0f);
		vMatrixSlice.rotate( 90.0f, 0.0f, 1.0f, 0.0f);
		break;
	case 1:	// Y
		vMatrixSlice.rotate( 90.0f, 1.0f, 0.0f, 0.0f);
		vMatrixSlice.rotate(180.0f, 0.0f, 1.0f, 0.0f);
		break;
	case 2: // Z
		vMatrixSlice.rotate(180.0f, 0.0f, 0.0f, 1.0f);
		;
	}

	updateAndRender();
}

void RenderWindow::mouseReleaseEvent(QMouseEvent *event)
{
	isMouseMoving = false;

	updateAndRender();
}

void RenderWindow::mousePressEvent(QMouseEvent *event)
{
	isMouseMoving = true;
	lastMousePos = event->pos();

	updateAndRender();
}

void RenderWindow::mouseMoveEvent(QMouseEvent *event)
{
	if(!isMouseMoving)
		return;

	int dx = event->x() - lastMousePos.x();
	int dy = event->y() - lastMousePos.y();

	if(event->buttons() == Qt::LeftButton)
	{
		alphaView -= dx;
		while(alphaView < 0.0f)
			alphaView += 360.0f;
		while(alphaView >= 360.0f)
			alphaView -= 360.0f;
		
		betaView -= dy ;
		while(betaView < 0.0f)
			betaView += 360.0f;
		while(betaView >= 360.0f)
			betaView -= 360.0f;

		//if(dx != 0.0f)
		//	azimuthView += qAtan(dy / dx);
		//
		//while(azimuthView < -180.0f)
		//	azimuthView += 360.0f;
		//while(azimuthView >= 180.0f)
		//	azimuthView -= 360.0f;
		//
		//while(inclinationView < -90.0f)
		//	inclinationView += 180.0f;
		//while(inclinationView >= 90.0f)
		//	inclinationView -= 180.0f;

		//QVector3D dPosVolume;
		//dPosVolume.setX(dx);
		//dPosVolume.setY(dy);
		//dPosVolume.setZ(qSqrt(qPow(distanceView, 2.0f)
		//	- qPow(dPosVolume.x(), 2.0f)
		//	- qPow(dPosVolume.y(), 2.0f)));
		//dPosVolume.setZ(0.0f);
		//
		//viewPosVolume += vMatrixVolume * dPosVolume;
		//
		//vMatrixVolume.setToIdentity();
		//vMatrixVolume.lookAt(viewPosVolume,
		//	QVector3D(0.0f, 0.0f, 0.0f),
		//	QVector3D(0.0f, 1.0f, 0.0f));
	}
	else if(event->buttons() == Qt::RightButton)
	{
		//mMatrixVolume.rotate(dx, 0.0f, 1.0f, 0.0f);
		//mMatrixVolume.rotate(dy, 1.0f, 0.0f, 0.0f);
	}

	lastMousePos = event->pos();
	event->accept();

	updateAndRender();
}

void RenderWindow::wheelEvent(QWheelEvent *event)
{
	int delta = event->delta();

	if(event->orientation() == Qt::Vertical)
	{
		if(delta < 0.0f)
			distanceView *= 1.1f;
		else if(delta > 0.0f)
			distanceView *= 0.9f;

		updateAndRender();
	}

	event->accept();
}

void RenderWindow::setRenderType(int _renderType)
{
	if(_renderType == RENDER_VOLUME)
		renderType = RENDER_VOLUME;
	else if(_renderType == RENDER_SLICE)
		renderType = RENDER_SLICE;

	updateAndRender();
}

void RenderWindow::setIsPerspectiveProjection(bool _perspectiveProjection)
{
	if(_perspectiveProjection)
		projectionType = PERSPECTIVE;
	else
		projectionType = ORTHOGONAL;

	updateAndRender();
}

void RenderWindow::setRenderShaded(bool _renderShaded)
{
	volumeShadedRendering = _renderShaded;

	updateAndRender();
}

void RenderWindow::setVolumeStepSize(double _stepSize)
{
	volumeStepsize = (float) _stepSize;
	updateAndRender();
}

void RenderWindow::setVolumeRenderMode(int _volumeRenderMode)
{
	if(_volumeRenderMode == COMPOSITION)
		volumeRenderMode = COMPOSITION;
	else if(_volumeRenderMode == AVERAGING)
		volumeRenderMode = AVERAGING;
	else if(_volumeRenderMode == MAXIMUM_INTENSITY_PROJECTION)
		volumeRenderMode = MAXIMUM_INTENSITY_PROJECTION;

	updateAndRender();
}

void RenderWindow::setVolumeCompositionMode(bool _volumeCompositionMode)
{
	if(_volumeCompositionMode == FRONT_TO_BACK)
		volumeCompositionMode = FRONT_TO_BACK;
	else if(_volumeCompositionMode == BACK_TO_FRONT)
		volumeCompositionMode = BACK_TO_FRONT;

	updateAndRender();
}

void RenderWindow::setRenderCannyEdges(bool _showCannyEdges)
{
	showCannyEdges = _showCannyEdges;
	render();
}

void RenderWindow::setRenderJunctions(bool _showJunctions)
{
	showJunctions = _showJunctions;
	render();
}

void RenderWindow::setRenderXJunctions(bool _showXJunctions)
{
	showXJunctions = _showXJunctions;
	render();
}

void RenderWindow::setUseBlurredTF(bool _useBlurredTF)
{
	useBluredTF = _useBlurredTF;
	updateAndRender();
	//setTransferfunction(copyTransferfunctionValues);
}

void RenderWindow::initializeMatrices()
{
	alphaView = 0.0f;
	betaView = 0.0f;
	distanceView = 2.5f;

	updateAndRender();
}

void RenderWindow::initializeSettings()
{
	// set initial values for variables

	volumeRenderMode = COMPOSITION;
	projectionType = PERSPECTIVE;
	volumeCompositionMode = FRONT_TO_BACK;
	volumeRenderMode = COMPOSITION;
	volumeStepsize = 0.005f;
	volumeShadedRendering = false;
	showCannyEdges = false;
	showJunctions = false;
	showXJunctions = false;
	useBluredTF = false;
	sigmaTF = 2.0f;
	filterSizeTF = 15;

	isMouseMoving = false;

	setSliceDirection(2);
	sliceNumber = 0.5f;

	distanceView = 2.0f;
	alphaView = 0.0f;
	betaView = 0.0f;
	inclinationView = 0.0f;
	azimuthView = 0.0f;
	viewPosVolume.setX(0.0f); viewPosVolume.setY(0.0f); viewPosVolume.setZ(distanceView);
}