#ifndef RENDERWINDOW_H
#define RENDERWINDOW_H

#include <QWindow>
#include <QtOpenGL>
#include <QOpenGLFunctions_3_2_Core>
#include <QTimer>

#include "Volume.h"
#include "cannyedge.h"

#include "TransferWidget.hpp"

#include "Tools.h"

/**
 * Main rendering window.
 * The main rendering window, which is on the one hand used to do all the OpenGL rendering and on the other hand to show additional stuff.
 * @see QWindow
 * @author Christian Hirsch
 * @author Clemens Roegner
 */
class RenderWindow : public QWindow
{
	Q_OBJECT

public:

	/**
	 * Constructor.
	 * Creates a new instance of the RenderWindow class.
	 * @param screen The screen used to display.
	 */
	RenderWindow(QScreen *screen = 0);

	/**
	 * Destructor.
	 */
	~RenderWindow();

	/**
	 * Initialization method.
	 * Initializes all the necessary objects, settings etc. that are used for the (OpenGL) rendering.
	 */
	void initialize();

	/**
	 * Updates the settings.
	 * Updates the settings, should be called when viewing, lightning etc. has changed.
	 */
	void update();

	/**
	 * Renders the image.
	 * Renders the image to the screen.
	 */
	void render();

	/**
	 * Updates the settings and renders the image.
	 * @see update
	 * @see render
	 */
	void updateAndRender();

	/**
	 * Resize the window.
	 * This method needs to be called when the window size changes (framebuffers/textures need to be reallocated etc.)
	 * @param width New width of the window.
	 * @param height new height of the window.
	 */
	void resize(int width, int height);

	/**
	 * Sets volume data.
	 * Loads new volume data from an given path.
	 * @param strFilename Path to the new volume data file.
	 */
	void setVolumeData(const std::string & strFilename);

	/**
	 * Sets the transfer function.
	 * Creates a new texture which stores the interpolated transferfunction values with the given control points.
	 * @param features The control points of the transferfunction.
	 */
	void setTransferfunction(const vector<TransferWidget::TransferFeature> &features);

	/**
	 * Initializes the matrices for OpenGL (viewing, projection, model matrix).
	 */
	void initializeMatrices();
	/**
	 * Initializes the settings for the whole volume rendering.
	 */
	void initializeSettings();

	/**
	 Render types.
	 */
	enum RenderType
	{
		RENDER_SLICE = 0,		/**< Render single slices of the volume. */
		RENDER_VOLUME			/**< Do 3D volume rendering. */
	};

	/**
	 * Projection types.
	 */
	enum ProjectionType
	{
		ORTHOGONAL = 0,			/**< Use orthogonal projection for 3D rendering. */
		PERSPECTIVE				/**< Use perspective projection for 3D rendering. */
	};

	/**
	 * Volume rendering composition modes.
	 */
	enum VolumeCompositionMode
	{
		FRONT_TO_BACK = true,	/**< Use front to back composition. */
		BACK_TO_FRONT = false	/**< Use back to front composition. */
	};

	/**
	 * Volume rendering modes.
	 */
	enum VolumeRenderMode
	{
		COMPOSITION = 0,		/**< Use composition of volume intensities. */
		AVERAGING,				/**< Use averaging of volume intensities. */
		MAXIMUM_INTENSITY_PROJECTION	/**< Use maximum intensity projection. */
	};

public slots:
	/**
	 * Log OpenGL messages.
	 * @param message QOpenGLDebugMessage that should be logged.
	 */
	void onMessageLogged( QOpenGLDebugMessage message );
	/**
	 * If timer is used, this slot should be triggered.
	 */
	void onTimeout();
	/**
	 * Renders a slice.
	 * @param _value Number of slice, which should be rendered.
	 */
	void updateSlice(int _value);
	/**
	 * Sets the new slice direction.
	 * @param _dir The new direction which should be rendered. 0 = x, 1 = y, 2 = z
	 */
	void setSliceDirection(int _dir);
	/**
	 * Sets a new render type.
	 * @param _renderType The new render type to be used.
	 * @see RenderType
	 */
	void setRenderType(int _renderType);
	/**
	 * Sets the projection type.
	 * @param _perspectiveProjection True if perspective projection to use.
	 * @see ProjectionType
	 */
	void setIsPerspectiveProjection(bool _perspectiveProjection);
	/**
	 * Sets the shading mode.
	 * @param _renderShaded True if shaded rendering should be used.
	 */
	void setRenderShaded(bool _renderShaded);
	/**
	 * Sets the new step size for scanning 3D volume.
	 * @param _stepSize The step size to use for scanning the volume.
	 */
	void setVolumeStepSize(double _stepSize);
	/**
	 * Sets the new volume render mode.
	 * @param _volumeRenderMode The new volume render mode.
	 * @see VolumeRenderMode
	 */
	void setVolumeRenderMode(int _volumeRenderMode);
	/**
	 * Sets the new volume composition mode.
	 * @param _compositingMode The new volume composition mode.
	 * @see VolumeCompositionMode
	 */
	void setVolumeCompositionMode(bool _compositingMode);
	/**
	 * Sets the rendering of the canny edges
	 * @param _showCannyEdges True if canny edges to render.
	 */
	void setRenderCannyEdges(bool _showCannyEdges);
	/**
	 * Sets the rendering of the junctions.
	 * @param _showJunctions True if junctions to render.
	 */
	void setRenderJunctions(bool _showJunctions);
	/**
	 * Sets the rendering of the X-junctions.
	 * @param _showXJunctions True if X junctions to render.
	 */
	void setRenderXJunctions(bool _showXJunctions);
	/**
	 * Sets the use of gaussian transferfunction.
	 * @param _useBlurredTF True if gaussian (blurred) transferfunctions to use.
	 */
	void setUseBlurredTF(bool _useBlurredTF);
	/**
	 * Calculate the depth order (energy value).
	 */
	void calculateDepthEnergy();

signals:
	/**
	 * Is emitted when depth order (energy value) has changed.
	 */
	void depthOrderChanged(float);

private:
	void prepareShaders();
	void prepareTextures();
	void prepareFramebuffers();
	void prepareQuadVertexArrayObject();
	void prepareCubeVertexArrayObject();

	void updateSlice();
	void updateVolume();
	void renderSlice();
	void renderVolume();
	void renderCubeFrontfaces();
	void renderCubeBackfaces();

	void renderTextureToScreen(GLuint texture, bool clearScreen = true);
	void renderTextureRectToScreen(GLuint texture, bool clearScreen = true);
	void renderQImageToScreen(const QImage &image, bool clearScreen = true);

	CannyEdge *cannyEdgeDetector;

	QOpenGLContext *v_context;

	QOpenGLFunctions_3_2_Core *v_funcs;

	QOpenGLShaderProgram *v_sliceShader;
	QOpenGLShaderProgram *v_cubeShader;
	QOpenGLShaderProgram *v_rayShader;
	QOpenGLShaderProgram *v_imageRectShader;
	QOpenGLShaderProgram *v_imageShader;
	
	QOpenGLFramebufferObject *v_frontFB;
	QOpenGLFramebufferObject *v_backFB;
	QOpenGLFramebufferObject *v_screenRectFB;
	
	QOpenGLVertexArrayObject v_quad;
	QOpenGLVertexArrayObject v_cube;

	QOpenGLDebugLogger *m_logger;

	float *verticesFullscreenQuad;

	GLuint v_transferfunctionTex;
	GLuint v_tmpTex;
	QImage junctionImage;
	QImage xJunctionImage;
	
	QMatrix4x4 vMatrixSlice;
	float sliceNumber;
	int sliceDirection;
	GLuint v_sliceTex;

	Volume *volumeData;
	QMatrix4x4 pMatrixVolume;
	QMatrix4x4 vMatrixVolume;
	QMatrix4x4 mMatrixVolume;
	VolumeRenderMode volumeRenderMode;
	VolumeCompositionMode volumeCompositionMode;
	float volumeStepsize;
	bool volumeShadedRendering;
	bool showCannyEdges;
	bool showJunctions;
	bool showXJunctions;
	bool useBluredTF;

	QMatrix4x4 mvpMatrixScreenRect;

	RenderType renderType;
	ProjectionType projectionType;

	float alphaView;
	float betaView;
	float distanceView;
	float inclinationView;
	float azimuthView;
	QVector3D viewPosVolume;
	QPoint lastMousePos;
	bool isMouseMoving;

	float depthOrdering;

	float *copyTransferfunctionValues;
	float sigmaTF;
	int filterSizeTF;

protected:
	/**
	 * Saves the current mouse position, for handling rotation.
	 * @param event Event which calls this funtion.
	 */
	void mousePressEvent(QMouseEvent *event);
	/**
	 * Sets variables needed for rotation.
	 * @param event Event which calls this function.
	 */
    void mouseMoveEvent(QMouseEvent *event);
	/**
	 * Documents when a mouse button has been released.
	 * @param event Event which calls this funtion.
	 */
	void mouseReleaseEvent(QMouseEvent *event);
	/**
	 * Is used to handle zooming.
	 * @param event Event which calls this funtion.
	 */
	void wheelEvent(QWheelEvent *event);
};

#endif // RENDERWINDOW_H
