#include <QtGui>
#include <stdio.h>

#include "vis2.hpp"

Vis2::Vis2(QWidget *parent)
	: QWidget(parent)
{
	renderWindow.initialize();

	createMenu();
	createGLGroupBox();
	createFileGroupBox();

	createSliceSettingsTab();
	createVolumeSettingsTab();
	createSettingsTabWidget();

	createTransferGroupBox();

	QWidget *leftSide = new QWidget;
	QVBoxLayout *leftLayout = new QVBoxLayout;
	leftLayout->addWidget(settingsTabWidget);
	leftLayout->addWidget(fileGroupBox);
	leftSide->setLayout(leftLayout);

	QVBoxLayout *mainLayout = new QVBoxLayout;
	mainLayout->addWidget(leftSide);
	mainLayout->addWidget(transferGroupBox);
	mainLayout->setMenuBar(menuBar);

	setLayout(mainLayout);

	setWindowTitle(tr("Volume Visualization"));
}

void Vis2::openVolume()
{
	volumeFileLocation->setText(QFileDialog::getOpenFileName(this));

	if(volumeFileLocation->text().isEmpty())
		return;

	renderWindow.setVolumeData(volumeFileLocation->text().toStdString());
}

void Vis2::openProject()
{
	projFileLocation->setText(QFileDialog::getOpenFileName(this));

	if(projFileLocation->text().isEmpty())
		return;

	FILE *project;

	project = fopen(projFileLocation->text().toLatin1(), "r");

	tWidget->clear();

	char buf[2048];

	while(fgets(buf, 2048, project) != NULL)
	{
		char *ptr;

		ptr = strstr(buf, "=");
		if(ptr != 0)
		{
			volumeFileLocation->setText(ptr);
			volumeFileLocation->setText(volumeFileLocation->text().remove(0,1).trimmed());
			if(volumeFileLocation->text().isEmpty())
				continue;
			renderWindow.setVolumeData(volumeFileLocation->text().toStdString());
		}
		else
		{
			float r, g, b, x, y;

			r = (float) strtod(strtok(buf, "/"), NULL);
			g = (float) strtod(strtok(0, "/"), NULL);
			b = (float) strtod(strtok(0, "/"), NULL);
			x = (float) strtod(strtok(0, "/"), NULL);
			y = (float) strtod(strtok(0, "/"), NULL);

			tWidget->setNewRGBXY(r, g, b, x, y);
		}
	}
	tWidget->resizeEvent(NULL);
	fclose(project);
}

void Vis2::saveProject()
{
	projFileLocation->setText(QFileDialog::getSaveFileName(this));

	if(projFileLocation->text().isEmpty())
		return;

	FILE *project;

	std::vector<TransferWidget::TransferFeature> features =
		tWidget->getTransferControlValues();

	project = fopen(projFileLocation->text().toLatin1(), "w");

	if(!projTransferOnly->isChecked()
		&& !volumeFileLocation->text().isEmpty())
	{
		fprintf(project, "volume=");
		fprintf(project, volumeFileLocation->text().toLatin1());
		fprintf(project, "\n");
	}

	for(int i=0; i < features.size(); i++)
	{
		fprintf(project, "%f/%f/%f/%f/%f\n",
			features[i].color.redF(),
			features[i].color.greenF(),
			features[i].color.blueF(),
			features[i].pos,
			1.0f - features[i].alpha);
	}

	fclose(project);
}

void Vis2::saveScreenshot()
{

}

void Vis2::createMenu()
{
	menuBar = new QMenuBar;

	fileMenu = new QMenu(tr("&File"));
	exitAction = fileMenu->addAction(tr("&Exit"));
	menuBar->addMenu(fileMenu);

	connect(exitAction, SIGNAL(triggered()), this, SLOT(close()));
}

void Vis2::createGLGroupBox()
{
}

void Vis2::createFileGroupBox()
{
	fileGroupBox = new QGroupBox(tr("File"));
	fileGroupBox->setMaximumWidth(350);

	QFormLayout *layout = new QFormLayout;

	volumeFileLocation = new QLineEdit;
	volumeFileOpenBtn = new QPushButton(tr("Open volume data..."));
	connect(volumeFileOpenBtn, SIGNAL(clicked()), this, SLOT(openVolume()));
	layout->addRow(new QLabel(tr("Volume data:")), volumeFileLocation);
	layout->addRow(new QLabel(tr("")), volumeFileOpenBtn);

	projFileLocation = new QLineEdit;
	projFileOpenBtn = new QPushButton(tr("Open project/transfer function..."));
	connect(projFileOpenBtn, SIGNAL(clicked()), this, SLOT(openProject()));
	projFileSaveBtn = new QPushButton(tr("Save project/transfer function..."));
	connect(projFileSaveBtn, SIGNAL(clicked()), this, SLOT(saveProject()));
	projTransferOnly = new QCheckBox(tr("Save transfer function only"));
	layout->addRow(new QLabel(tr("Project data:")), projFileLocation);
	layout->addRow(new QLabel(tr("")), projTransferOnly);
	layout->addRow(new QLabel(tr("")), projFileOpenBtn);
	layout->addRow(new QLabel(tr("")), projFileSaveBtn);

	fileGroupBox->setLayout(layout);
}

void Vis2::createSettingsTabWidget()
{
	settingsTabWidget = new QTabWidget;
	settingsTabWidget->setMaximumWidth(350);

	settingsTabWidget->addTab(sliceSettingsTab, tr("Slice rendering"));
	settingsTabWidget->addTab(volumeSettingsTab, tr("Volume rendering"));

	connect(settingsTabWidget, SIGNAL(currentChanged(int)), &renderWindow, SLOT(setRenderType(int)));
}

void Vis2::createSliceSettingsTab()
{
	sliceSettingsTab = new QWidget();

	QFormLayout *layout = new QFormLayout;
	sliceSettingsTab->setLayout(layout);

	xRadioBtn = new QRadioButton("X-direction", this);
	connect(xRadioBtn, SIGNAL(toggled(bool)), this, SLOT(directionChanged(bool)));

	yRadioBtn = new QRadioButton("Y-direction", this);
	connect(yRadioBtn, SIGNAL(toggled(bool)), this, SLOT(directionChanged(bool)));

	zRadioBtn = new QRadioButton("Z-direction", this);
	connect(zRadioBtn, SIGNAL(toggled(bool)), this, SLOT(directionChanged(bool)));
	zRadioBtn->setChecked(true);

	layout->addRow(new QLabel("Show slices in:"), xRadioBtn);
	layout->addRow(new QLabel(""), yRadioBtn);
	layout->addRow(new QLabel(""), zRadioBtn);

	connect(this, SIGNAL(directionChangedTo(int)), &renderWindow, SLOT(setSliceDirection(int)));

	sliceSlider = new QSlider(Qt::Horizontal, this);
	sliceSlider->setMinimum(0);
	sliceSlider->setMaximum(1023);
	sliceSlider->setSingleStep(1);
	sliceSlider->setValue(1023/2);
	connect(sliceSlider, SIGNAL(valueChanged(int)), &renderWindow, SLOT(updateSlice(int)));
	layout->addRow(new QLabel(tr("Select slice:")), sliceSlider);
}

void Vis2::createVolumeSettingsTab()
{
	volumeSettingsTab = new QWidget();

	QFormLayout *layout = new QFormLayout;
	volumeSettingsTab->setLayout(layout);

	// shading
	lightCheckBox = new QCheckBox;
	connect(lightCheckBox, SIGNAL(toggled(bool)),
		&renderWindow, SLOT(setRenderShaded(bool)));
	lightCheckBox->setChecked(false);
	layout->addRow(new QLabel(tr("Shading:")), lightCheckBox);

	// perspective projection
	perspProjCheckBox = new QCheckBox;
	connect(perspProjCheckBox, SIGNAL(toggled(bool)),
		&renderWindow, SLOT(setIsPerspectiveProjection(bool)));
	perspProjCheckBox->setChecked(true);
	layout->addRow(new QLabel(tr("Perspective projection:")),
		perspProjCheckBox);

	// step size
	stepsizeSpinBox = new QDoubleSpinBox;
	stepsizeSpinBox->setMinimum(0.001);
	stepsizeSpinBox->setSingleStep(0.001);
	stepsizeSpinBox->setDecimals(3);
	connect(stepsizeSpinBox, SIGNAL(valueChanged(double)),
		&renderWindow, SLOT(setVolumeStepSize(double)));
	stepsizeSpinBox->setValue(0.005);
	layout->addRow(new QLabel(tr("Step size:")), stepsizeSpinBox);

	// render mode
	modeBox = new QComboBox;
	modeBox->addItem(tr("Compositing"));
	modeBox->addItem(tr("Maximum intensity projection"));
	modeBox->addItem(tr("Averaging"));
	connect(modeBox, SIGNAL(currentIndexChanged(int)),
		&renderWindow, SLOT(setVolumeRenderMode(int)));
	layout->addRow(new QLabel(tr("Render mode:")), modeBox);

	// compositing mode
	f2bRadioButton = new QRadioButton(tr("Front to Back"));
	f2bRadioButton->setChecked(true);
	connect(f2bRadioButton, SIGNAL(toggled(bool)),
		&renderWindow, SLOT(setVolumeCompositingMode(bool)));
	b2fRadioButton = new QRadioButton(tr("Back to Front"));
	layout->addRow(new QLabel(tr("Compositing mode:")), f2bRadioButton);
	layout->addRow(new QLabel(tr("")), b2fRadioButton);
	f2bRadioButton->setChecked(true);

	// Canny Edges
	showCannyEdgesCheckBox = new QCheckBox;
	connect(showCannyEdgesCheckBox, SIGNAL(toggled(bool)),
		&renderWindow, SLOT(setRenderCannyEdges(bool)));
	showCannyEdgesCheckBox->setChecked(false);
	layout->addRow(new QLabel(tr("Show Canny Edges:")), showCannyEdgesCheckBox);

	// Junctions
	showJunctionsCheckBox = new QCheckBox;
	connect(showJunctionsCheckBox, SIGNAL(toggled(bool)),
		&renderWindow, SLOT(setRenderJunctions(bool)));
	showJunctionsCheckBox->setChecked(false);
	layout->addRow(new QLabel(tr("Show Junctions:")), showJunctionsCheckBox);

	// X Junctions
	showXJunctionsCheckBox = new QCheckBox;
	connect(showXJunctionsCheckBox, SIGNAL(toggled(bool)),
		&renderWindow, SLOT(setRenderXJunctions(bool)));
	showXJunctionsCheckBox->setChecked(false);
	layout->addRow(new QLabel(tr("Show X-Junctions:")), showXJunctionsCheckBox);

	// Blurred TF
	useBlurredTFCheckBox = new QCheckBox;
	connect(useBlurredTFCheckBox, SIGNAL(toggled(bool)),
		&renderWindow, SLOT(setUseBlurredTF(bool)));
	useBlurredTFCheckBox->setChecked(false);
	layout->addRow(new QLabel(tr("Use Blurred TF:")), useBlurredTFCheckBox);

	// depth ordering label
	depthOrderingLabel = new QLabel(tr("not calculated"));
	connect(&renderWindow, SIGNAL(depthOrderChanged(float)),
		this, SLOT(updateDepthOrder(float)));
	layout->addRow(new QLabel(tr("Depth ordering:")), depthOrderingLabel);

	// calculate depth energy
	calculateDepthEnergyButton = new QPushButton("Calculate depth energy");
	connect(calculateDepthEnergyButton, SIGNAL(clicked()),
		&renderWindow, SLOT(calculateDepthEnergy()));
	layout->addRow(calculateDepthEnergyButton);

	// reset Button
	resetButton = new QPushButton(tr("Reset settings"));
	connect(resetButton, SIGNAL(clicked()),
		this, SLOT(resetSettings()));
	layout->addRow(resetButton);
}

void Vis2::changeSliderRange(int min, int max)
{
	sliceSlider->setMinimum(min);
	sliceSlider->setMaximum(max);
	sliceSlider->setValue((max-min)/2 + min);
}

void Vis2::createTransferGroupBox()
{
	transferGroupBox = new QGroupBox(tr("Transfer function"));
	transferGroupBox->setMaximumHeight(200);
	QHBoxLayout *layout = new QHBoxLayout;
	transferGroupBox->setLayout(layout);

	tWidget = new TransferWidget(this);
	connect(tWidget, SIGNAL(transferChanged()), this, SLOT(transferChanged()));
	layout->addWidget(tWidget);
}

void Vis2::directionChanged(bool checked)
{
	if(!checked)
		return;

	if(xRadioBtn->isChecked())
	{
		emit directionChangedTo(0);
		return;
	}

	if(yRadioBtn->isChecked())
	{
		emit directionChangedTo(1);
		return;
	}

	emit directionChangedTo(2);
}

void Vis2::transferChanged()
{
	renderWindow.setTransferfunction(tWidget->getTransferControlValues());
}

void Vis2::resetSettings()
{
	lightCheckBox->setChecked(false);
	perspProjCheckBox->setChecked(true);
	modeBox->setCurrentIndex(0);
	stepsizeSpinBox->setValue(0.005);
	f2bRadioButton->setChecked(true);
	showCannyEdgesCheckBox->setChecked(false);
	useBlurredTFCheckBox->setChecked(false);
	renderWindow.initializeMatrices();
}

void Vis2::updateDepthOrder(float depthOrder)
{
	depthOrderingLabel->setText(QString::number(depthOrder));
}