#ifndef VIS2_H
#define VIS2_H

#include "TransferWidget.hpp"
#include "renderwindow.h"

 class QAction;
 class QDialogButtonBox;
 class QGroupBox;
 class QLabel;
 class QLineEdit;
 class QMenu;
 class QMenuBar;
 class QPushButton;
 class QTextEdit;
 class QRadioButton;
 class QFileDialog;
 class QSlider;

/**
 * Main application widget.
 * This widget is the main application widget, which contains all UI components
 * for visualizing volume data.
 * @author Christian Hirsch
 * @author Clemens Roegner
 */
class Vis2 : public QWidget
{
    Q_OBJECT

public:
	/**
	 * Constructer.
	 * Creates a new widget. Creats all needed UI components.
	 * @see QWidget.
	 */
    Vis2(QWidget *parent = 0);

private:
	/**
	 * Creates the menu.
	 * This function creates a new file-menu.
	 */
    void createMenu();
	/**
	 * Creates the goup box for the OpenGL window.
	 */
    void createGLGroupBox();
	/**
	 * Creates the group box for handling volume, project/transfer function
	 * and screenshot files.
	 */
    void createFileGroupBox();
	/**
	 * Creates the group box, which contains the widget for the transfer function.
	 */
	void createTransferGroupBox();

	/**
	 * Creates a tab, which contains settings for handling the slice/volume
	 * rendering.
	 */
	void createSettingsTabWidget();
	/**
	 * Creates the UI components for handling the slice rendering.
	 */
	void createSliceSettingsTab();
	/**
	 * Creates the UI components for handling the volume rendering.
	 */
	void createVolumeSettingsTab();

    QMenuBar *menuBar;

    QGroupBox *glGroupBox;
    QGroupBox *fileGroupBox;
	QGroupBox *transferGroupBox;

	QTabWidget *settingsTabWidget;
	QWidget *sliceSettingsTab;
	QWidget *volumeSettingsTab;

    QMenu *fileMenu;
    QAction *exitAction;
	
    // slice settings
    QRadioButton *zRadioBtn;
    QRadioButton *yRadioBtn;
    QRadioButton *xRadioBtn;
    QSlider *sliceSlider;

	// volume settings
	QCheckBox *lightCheckBox;
	QCheckBox *perspProjCheckBox;
	QComboBox *modeBox;
	QDoubleSpinBox *stepsizeSpinBox;
	QRadioButton *f2bRadioButton;
	QRadioButton *b2fRadioButton;
	QSlider *xPlaneSlider;
	QSlider *yPlaneSlider;
	QSlider *zPlaneSlider;
	QPushButton *resetButton;
	QPushButton *calculateDepthEnergyButton;
	QCheckBox *showCannyEdgesCheckBox;
	QCheckBox *showJunctionsCheckBox;
	QCheckBox *showXJunctionsCheckBox;
	QCheckBox *useBlurredTFCheckBox;
	QLabel *depthOrderingLabel;

    // transfer function
    TransferWidget *tWidget;

    // file loading
    QLineEdit *volumeFileLocation;
    QPushButton *volumeFileOpenBtn;
    QAction *volumeFileOpenAct;
	QLineEdit *projFileLocation;
    QPushButton *projFileOpenBtn;
	QPushButton *projFileSaveBtn;
	QCheckBox *projTransferOnly;
	QPushButton *screenshotSaveBtn;
    QAction *projFileOpenAct;

	RenderWindow renderWindow;

    // transfer function

public slots:
	/**
	 * Function when range of slices is changed.
	 * This function changes the range of the slice-slider.
	 * When the range changes, the slice slider is set to the middle of the range.
	 */
	void changeSliderRange(int min, int max);
	/**
	 * Resets settings for volume rendering.
	 * This function resets the setting for the volume rendering.
	 */
	void resetSettings();
	/**
	 * Updates the depth order.
	 * This solt updates the depth order for the depth order label.
	 */
	void updateDepthOrder(float depthOrder);

private slots:
	/**
	 * Function for opening a volume.
	 */
    void openVolume();
	/**
	 * Function for opening a project or transferfunction file.
	 * If a project file will be opened, it also loads the volume file.
	 */
	void openProject();
	/**
	 * Function for saving a project or transferfunction file.
	 */
	void saveProject();
	/**
	 * Function for saving the current OpenGL rendering to a png file.
	 */
	void saveScreenshot();
	/**
	 * Function for handling the changes of the slice direction radio buttons.
	 */
    void directionChanged(bool);
	/**
	 * Function for handling the change of the transfer function.
	 */
    void transferChanged();

signals:
	/**
	 * Signal, which is emitted, when the direction of the slices has changed.
	 */
	void directionChangedTo(int);
};

#endif // Vis2_H
